/* KeyboardHost: Copyright 2011 John Elliott <jce@seasip.demon.co.uk>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package info.seasip.keyboardhost;

import javax.swing.SwingUtilities;

/** HostPanel emulates the hardware the emulated keyboard is talking to.
 * Rather than do this by running a real ROM (say, IBM 1503033) under
 * emulation, it does hex encoding and decoding in Java, in collaboration
 * with the KeyboardPCB class.
 *
 * @author jelliott
 */
public class HostPanel extends javax.swing.JPanel
{
    private HostListener listener = null;

    /** Creates new form HostPanel */
    public HostPanel() {
        initComponents();
    }

    public void setListener(HostListener l)
    {
        listener = l;
    }

    /** Called wneh a hex byte is received from the emulated keyboard.
     *
     * @param b The byte received.
     * @param par The accompanying parity bit.
     */
    public void receiveHex(byte b, boolean par)
    {
        int n;
        boolean p = true;
        int mask = 0x80;

        // Calculate parity, and see if it matches what the keyboard sent.
        for (n = 0; n < 8; n++)
        {
           if ((b & mask) != 0) p = !p;
           mask = mask >> 1;
        }
        String s = receivedData.getText();

        // Append the new byte, followed by space (if parity is good) or star
        // (if parity is bad).
        s += String.format("%02x%c", b, (par == p) ? ' ' : '*');
        receivedData.setText(s);
        receivedData.select(s.length() - 3, s.length());
    }

    /** Clear the display of hex bytes received.
     */
    public void clearBuffer()
    {
        receivedData.setText("");
    }


    private int recvState = 0;
    private boolean lastClock = true;
    private int incoming = 0;
    private boolean parity = false;

    /** Handle a transition on the 'clock' line.
     *
     * @param clock Current state of the clock line.
     * @param data Current state of the data line.
     */
    public void onClockChange(boolean clock, boolean data)
    {
        // On a low -> high transition, we read the data line and act
        // accordingly.
        if (lastClock == false && clock == true)    
        {
            switch (recvState)
            {
                // Expecting the start bit (always 0).
                case 0:
                    incoming = 0;
                    if (data == false) ++recvState; 
                    break;

                // Expecting the data byte, LSB first.
                case 1: case 2: case 3: case 4:
                case 5: case 6: case 7: case 8:
                    incoming = incoming >> 1;
                    if (data) incoming |= 0x80;
                    ++recvState;
                    break;

                // Expecting the parity bit
                case 9:
                    parity = data;
                    ++recvState;
                    break;

                // Expecting the stop bit (always 1).
                case 10:
                    if (data == true)
                    {
                        recvState = 0;
                        // Call receiveHex() on the UI thread.
                        Runnable r = new Runnable()
                        {
                            public void run()
                            {
                                receiveHex((byte)incoming, parity);
                            }
                        };
                        SwingUtilities.invokeLater(r);
                    }
                    break;
            }
        }
        lastClock = clock;
    }



    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        jScrollPane1 = new javax.swing.JScrollPane();
        jTextArea1 = new javax.swing.JTextArea();
        label1 = new java.awt.Label();
        jButton1 = new javax.swing.JButton();
        jScrollPane2 = new javax.swing.JScrollPane();
        receivedData = new javax.swing.JTextArea();
        sendData = new javax.swing.JTextField();

        jScrollPane1.setName("jScrollPane1"); // NOI18N

        jTextArea1.setColumns(20);
        jTextArea1.setRows(5);
        jTextArea1.setName("jTextArea1"); // NOI18N
        jScrollPane1.setViewportView(jTextArea1);

        setName("Form"); // NOI18N

        label1.setName("label1"); // NOI18N
        org.jdesktop.application.ResourceMap resourceMap = org.jdesktop.application.Application.getInstance(info.seasip.keyboardhost.KeyboardHostApp.class).getContext().getResourceMap(HostPanel.class);
        label1.setText(resourceMap.getString("label1.text")); // NOI18N

        jButton1.setText(resourceMap.getString("jButton1.text")); // NOI18N
        jButton1.setName("jButton1"); // NOI18N
        jButton1.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                sendHex(evt);
            }
        });

        jScrollPane2.setName("jScrollPane2"); // NOI18N

        receivedData.setColumns(20);
        receivedData.setEditable(false);
        receivedData.setLineWrap(true);
        receivedData.setRows(5);
        receivedData.setName("receivedData"); // NOI18N
        jScrollPane2.setViewportView(receivedData);

        sendData.setText(resourceMap.getString("sendData.text")); // NOI18N
        sendData.setName("sendData"); // NOI18N

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addContainerGap()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addComponent(jScrollPane2, javax.swing.GroupLayout.DEFAULT_SIZE, 409, Short.MAX_VALUE)
                    .addGroup(layout.createSequentialGroup()
                        .addComponent(label1, javax.swing.GroupLayout.PREFERRED_SIZE, 50, javax.swing.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(sendData, javax.swing.GroupLayout.PREFERRED_SIZE, 292, javax.swing.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                        .addComponent(jButton1)))
                .addContainerGap())
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addContainerGap()
                .addComponent(jScrollPane2, javax.swing.GroupLayout.PREFERRED_SIZE, 164, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING, false)
                    .addComponent(label1, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                    .addComponent(jButton1, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                    .addComponent(sendData, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addContainerGap(javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE))
        );
    }// </editor-fold>//GEN-END:initComponents

    private void sendHex(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_sendHex
        String s = sendData.getText();
        String [] contents = s.split("\\s");    // Split at spaces
        final byte [] parsed = new byte[contents.length];

        try
        {
            for (int n = 0; n < contents.length; n++)
            {
                parsed[n] = (byte)Short.parseShort(contents[n], 16);
            }
        }
        catch (NumberFormatException ex)
        {
            sendData.setText(ex.getMessage());
            return;
        }
        if (listener != null)
        {
            Runnable r = new Runnable()
            {
                public void run()
                {
                    for (int n = 0; n < parsed.length; n++)
                    {
                        listener.sendByte(parsed[n]);
                        try
                        {
                            Thread.sleep(500);
                        }
                        catch (Exception ex)
                        {

                        }
                    }
                }
            };
            new Thread(r).start();
        }
    }//GEN-LAST:event_sendHex


    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JButton jButton1;
    private javax.swing.JScrollPane jScrollPane1;
    private javax.swing.JScrollPane jScrollPane2;
    private javax.swing.JTextArea jTextArea1;
    private java.awt.Label label1;
    private javax.swing.JTextArea receivedData;
    private javax.swing.JTextField sendData;
    // End of variables declaration//GEN-END:variables

}
