/* KeyboardHost: Copyright 2011 John Elliott <jce@seasip.demon.co.uk>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package info.seasip.keyboardhost;

import java.util.LinkedList;

/** Emulates the keyboard printed circuit board. Interfaces the emulated
 * microcontroller to its UI (KeyboardPanel and HostPanel).
 *
 * @author John Elliott
 */
public class KeyboardPCB implements IOPort, HostListener
{
    private boolean hostClock = true;   // False if host is driving clock low
    private boolean hostData = true;    // False if host is driving data low
    private boolean kbClock = true;     // False if keyboard is driving clock low
    private boolean kbData = true;      // False if keyboard is driving data low
    private byte p1 = -1, p2 = -1, bus = -1;
    private byte keyToRead = -1;
    private byte sensitivity = -1;
    private boolean t1 = true;
    private LinkedList host2kb = new LinkedList();


    private KeyboardPanel keyboardPanel;
    private HostPanel hostPanel;

    public KeyboardPCB(KeyboardPanel kbp, HostPanel hp)
    {
        keyboardPanel = kbp;
        hostPanel = hp;
    }

    /** Add a byte to be sent to the keyboard by the host.
     *
     * @param b The byte to be sent.
     */
    public synchronized void sendByte(byte b)
    {
        host2kb.offer(b);
    }

    /** Are there bytes waiting to be sent by the host?
     *
     * @return true if there are bytes waiting, false if not.
     */
    public synchronized boolean bytesWaiting()
    {
        return host2kb.size() > 0;
    }

    /** Get the next byte that the host is sending, and remove from the queue.
     *
     * @return The byte.
     */
    public synchronized byte pullByte()
    {
        return (Byte)(host2kb.poll());
    }

    /** Get the next byte that the host is sending, leaving it in the queue.
     *
     * @return The byte.
     */
    public synchronized byte getByte()
    {
        return (Byte)(host2kb.peek());
    }


    private int receiveState = 0;
    private byte currentByte = 0;
    private boolean parity = false;

    /** See if the host has something to send to the keyboard. If it does,
     * convert the byte into the appropriate raises and lowers of the
     * keyboard's data line.
     */
    void checkHost()
    {
        byte mask;

        // If not in the middle of sending something, and nothing is
        // coming, float the data and clock lines.
        if (receiveState <= 0 && !bytesWaiting())
        {
            hostData = true;
            hostClock = true;
            return;
        }
        // Host does want to send something. Firstly, pull the clock line
        // low.
        switch (receiveState)
        {
            case 0: case 1: case 2: case 3:
                ++receiveState;
                hostClock = false;
                break;

            // Now release the clock so the keyboard can drive it, and send 
            // the start bit (always 0).
            case 4:
                currentByte = pullByte();
                parity = true;
                mask = 1;
                for (int n = 0; n < 8; n++)
                {
                    if ((currentByte & mask) != 0)
                    {
                        parity = !parity;
                    }
                    mask = (byte)(mask << 1);
                }
                ++receiveState;
                hostClock = true;
                hostData  = false;  // Start bit
                break;

            // Subsequent bits are transferred in onClockChange(), on a
            // low->high clock transition.
        }
        
    }




    private boolean lastClock = true;

    /** The clock line has changed.
     *
     */
    public void onClockChange()
    {
        // If the keyboard has changed its clock line and the host is not
        // driving it:
        if (lastClock != kbClock && hostClock == true)
        {
           // If host is not transmitting to keyboard, keyboard must be
           // transmitting to host. Hand off to the HostPanel.
           if (receiveState <= 0)
           {
               hostPanel.onClockChange(kbClock, kbData);
           }
           // Incoming data bits happen on a low->high transition
           else if (kbClock) switch(receiveState)
           {
               case 5:  case 6:   case 7:  case 8:
               case 9:  case 10:  case 11: case 12:
                       hostData = (currentByte & 1) != 0;
                       //System.out.printf("[%d] hostData=%d\n", receiveState, hostData ? 1 : 0);
                       currentByte = (byte)(currentByte >> 1);
                       ++receiveState;
                       break;
               case 13:
                       hostData = parity;
                       //System.out.printf("[%d] hostData=%d\n", receiveState, hostData ? 1 : 0);
                       ++receiveState;
                       break;
               case 14:
                       hostData = true; // Stop bit
                       receiveState++;
                       break;
               case 15:
                       if (kbData == false) // ACK
                       {
                           receiveState = 0;
                       }
                       break;
           }
        }
        lastClock = kbClock;
    }

    // Implementation of the IOPort interface.

    /** Read a microcontroller port.
     *
     * @param port 0 for the bus, 1-7 for numbered ports.
     * @return The state of the selected port.
     */
    public byte readPort(int port)
    {
        switch(port)
        {
            case 0: return readBus();
            case 1: return readP1();
            case 2: return readP2();
            default: return -1;
        }
    }

    public void writePort(int port, byte value)
    {
        switch(port)
        {
            case 0: writeBus(value); break;
            case 1: writeP1(value); break;
            case 2: writeP2(value); break;
        }
    }



    private void writeP1(byte value)
    {
        p1 = value;
    }

    // Reading P1 reads the keyboard ID.
    private byte readP1()
    {
        return (byte)((keyboardPanel.getID(false) << 2) | (p1 & 3));
    }

    // Writing P2 updates the clock and data lines, and the LEDs.
    private void writeP2(byte value)
    {
        p2 = value;
        kbClock = ((value & 0x40) == 0);        // Bit 6 drives clock line
        kbData  = ((value & 0x80) == 0);        // Bit 7 drives data line
        keyboardPanel.setLEDs((byte)(value & 7));
        onClockChange();
    }

    // Reading P2 reads the keyboard ID.
    private byte readP2()
    {
        return (byte)(((keyboardPanel.getID(true) & 0x3F) | 0xC0) & p2);
    }

    // The bus is connected to the capacitative key reader and the key matrix.
    public void writeBus(byte value)
    {
        bus = value;
        // The low 2 bits of P1 appear to control the capacitative key reader.
        switch (p1 & 3)
        {
            case 0: keyToRead   = (byte)(value & 0x7F); break;  // 00: Select row/col
            case 2: sensitivity =  (byte)(value & 3); break;    // 10: Set sensitivity
            case 3: t1 = !keyboardPanel.isKeyDown(keyToRead);   // 11: Read keyboard
                   // System.out.printf("Read key %02x\n", keyToRead);
                    break;
        }
    }

    public byte readBus()
    {
        return bus;
    }

    // The T0 line is connected (by means of a NAND gate) to the keyboard and
    // host data lines.
    public boolean readT0()
    {
        boolean result;
        checkHost();

        result = (!(hostData && kbData));  // If both are high, return low. Else return high.
        return result;
    }

    // The T1 line is connected to line 1 of P1, and the capacitative
    // scanning circuitry.
    public boolean readT1()
    {
        // If line 1 of P1 is low, this pulls T1 low.
        if ((p1 & 2) == 0) return false;

        // Otherwise return the state of the last keypress.
        return t1;
    }

    // The I line is connected to keyboard and host clock lines through a
    // NAND gate.
    public boolean readI()
    {
        checkHost();
        return (!(hostClock && kbClock));  // If both are high, return low. Else return high.
    }

    // There is no external memory.
    public byte readExtMem(int addr)
    {
        return -1;
    }

    public void writeExtMem(int addr, byte value)
    {

    }

}
