/* KeyboardHost: Copyright 2011 John Elliott <jce@seasip.demon.co.uk>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package info.seasip.keyboardhost;

import javax.swing.JToggleButton;
import javax.swing.JPanel;
import java.awt.GridBagLayout;
import java.awt.GridBagConstraints;
import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.SwingUtilities;

/** KeyboardPanel: Emulates the user-accessible parts of an IBM 3270PC
 * keyboard (with the addition of the LED panel from the IBM AT keyboard).
 *
 * @author John Elliott
 */
public class KeyboardPanel extends JPanel
{
    // 128 buttons for the 128 key positions.
    private JToggleButton [] buttons = new JToggleButton[128];

    // 12 checkboxes for the 12 identity jumpers.
    private JToggleButton [] keyid   = new JToggleButton[12];

    // 3 checkboxes for the 3 LEDs.
    private JToggleButton [] led     = new JToggleButton[3];

    // Layout is done AWT-style, rather than with NetBeans.
    private GridBagLayout layout = new GridBagLayout();

    // Current LED state.
    private int ledStatus = -1;

    // Keyboard ID (based on jumper settings).
    private byte kbID[] = new byte[2];

    // Captions for the LEDs.
    private static final String ledName[] =
    {
        "Num",
        "Caps",
        "Scrl"
    };

    // Positions for each of the 128 keys, in grid coordinates. Four ints for
    // each key: X, Y, width, height.
    private static final int keyPos[] =
    {
        0, 14, 2, 2,    // 00 L9
        0, 12, 2, 2,    // 01 L7
        0, 10, 2, 2,    // 02 L5
        0,  8, 2, 2,    // 03 L3
        0,  6, 2, 2,    // 04 L1
        2,  6, 2, 2,    // 05 L2
        8,  2, 2, 2,    // 06 PF1
        8,  0, 2, 2,    // 07 PF13
        2, 14, 2, 2,    // 08 L10
        2, 12, 2, 2,    // 09 L8
        2, 10, 2, 2,    // 0A L6
        2,  8, 2, 2,    // 0B L5
        6,  8, 3, 2,    // 0C Tab
        6,  6, 2, 2,    // 0D Backtick
        10, 2, 2, 2,    // 0E PF2
        10, 0, 2, 2,    // 0F PF14
        6, 14, 4, 2,    // 10 Ctrl
        6, 12, 3, 2,    // 11 Shift
        9, 12, 2, 2,    // 12 Backslash
        6, 10, 4, 2,    // 13 Caps Lock
        9,  8, 2, 2,    // 14 Q
        8,  6, 2, 2,    // 15 1
        12, 2, 2, 2,    // 16 PF3
        12, 0, 2, 2,    // 17 PF15
        12,14, 3, 2,    // 18 Alt
        11,12, 2, 2,    // 19 Z
        12,10, 2, 2,    // 1A S
        10,10, 2, 2,    // 1B A
        11, 8, 2, 2,    // 1C W
        10, 6, 2, 2,    // 1D 2
        14, 2, 2, 2,    // 1E PF4
        14, 0, 2, 2,    // 1F PF16
        15,12, 2, 2,    // 20 C
        13,12, 2, 2,    // 21 X
        14,10, 2, 2,    // 22 D
        13, 8, 2, 2,    // 23 E
        14, 6, 2, 2,    // 24 4
        12, 6, 2, 2,    // 25 3
        16, 2, 2, 2,    // 26 PF5
        16, 0, 2, 2,    // 27 PF17
        15,14,13, 2,    // 28 Space
        17,12, 2, 2,    // 29 V
        16,10, 2, 2,    // 2A F
        17, 8, 2, 2,    // 2B T
        15, 8, 2, 2,    // 2C R
        16, 6, 2, 2,    // 2D 5
        18, 2, 2, 2,    // 2E PF6
        18, 0, 2, 2,    // 2F PF18
        21,12, 2, 2,    // 30 N
        19,12, 2, 2,    // 31 B
        20,10, 2, 2,    // 32 H
        18,10, 2, 2,    // 33 G
        19, 8, 2, 2,    // 34 Y
        18, 6, 2, 2,    // 35 6
        20, 2, 2, 2,    // 36 PF7
        20, 0, 2, 2,    // 37 PF19
        28,14, 3, 2,    // 38 Right Alt
        23,12, 2, 2,    // 39 M
        22,10, 2, 2,    // 3A J
        21, 8, 2, 2,    // 3B U
        20, 6, 2, 2,    // 3C 7
        22, 6, 2, 2,    // 3D 8
        22, 2, 2, 2,    // 3E PF8
        22, 0, 2, 2,    // 3F PF20
        25,12, 2, 2,    // 40 ,
        24,10, 2, 2,    // 41 K
        23, 8, 2, 2,    // 42 I
        25, 8, 2, 2,    // 43 O
        26, 6, 2, 2,    // 44 0
        24, 6, 2, 2,    // 45 9
        24, 2, 2, 2,    // 46 PF9
        24, 0, 2, 2,    // 47 PF21
        27,12, 2, 2,    // 48 ,
        29,12, 2, 2,    // 49 /
        26,10, 2, 2,    // 4A L
        28,10, 2, 2,    // 4B ;
        27, 8, 2, 2,    // 4C P
        28, 6, 2, 2,    // 4D -
        26, 2, 2, 2,    // 4E PF10
        26, 0, 2, 2,    // 4F PF22
        31,14, 2, 2,    // 50 Dummy
        31,12, 2, 2,    // 51 Under right Shift
        30,10, 2, 2,    // 52 '
        32,10, 2, 2,    // 53 #
        29, 8, 2, 2,    // 54 [
        30, 6, 2, 2,    // 55 =
        28, 2, 2, 2,    // 56 PF11
        28, 0, 2, 2,    // 57 PF23
        33,14, 3, 2,    // 58 Right Ctrl
        33,12, 3, 2,    // 59 Right Shift
        34,10, 2, 2,    // 5A Return
        31, 8, 2, 2,    // 5B ]
        33, 8, 3, 2,    // 5C Other #
        32, 6, 2, 2,    // 5D Under backspace
        30, 2, 2, 2,    // 5E PF12
        30, 0, 2, 2,    // 5F PF24
        40, 14,2, 2,    // 60 Down
        38, 12,2, 2,    // 61 Left
        40, 12,2, 2,    // 62 Home
        40, 10,2, 2,    // 63 Up
        38,  8,2, 2,    // 64 Ins
        40,  8,2, 2,    // 65 PgDn
        34,  6,2, 2,    // 66 Backspace
        38,  6,2, 2,    // 67 PA1
        46, 14,2, 2,    // 68 Under keypad 0
        46, 12,2, 2,    // 69 Keypad 1
        42, 12,2, 2,    // 6A Right
        46, 10,2, 2,    // 6B Keypad 4
        46,  8,2, 2,    // 6C Keypad 7
        42,  8,2, 2,    // 6D Del
        40,  6,2, 2,    // 6E PA2
        42,  6,2, 2,    // 6F PA3
        48, 14,2, 2,    // 70 Keypad 0
        50, 14,2, 2,    // 71 Keypad .
        48, 12,2, 2,    // 72 Keypad 2
        48, 10,2, 2,    // 73 Keypad 5
        50, 10,2, 2,    // 74 Keypad 5
        48,  8,2, 2,    // 75 Keypad 8
        46,  6,2, 2,    // 76 Num lock
        48,  6,2, 2,    // 77 Keypad /
        52, 14,2, 2,    // 78 Under keypad enter
        52, 12,2, 2,    // 79 Keypad enter
        50, 12,2, 2,    // 7A Keypad 3
        52, 10,2, 2,    // 7B Under keypad +
        52,  8,2, 2,    // 7C Keypad +
        50,  8,2, 2,    // 7D Keypad 9
        50,  6,2, 2,    // 7E Keypad *
        52,  6,2, 2,    // 7F Keypad -
    };

    // Positions for spacer panels, placed between the different key groups.
    private static final int spacerPos[] =
    {
        4,  0,  2, 16,
        0,  4, 45,  2,
        36, 0,  2, 16,
        38, 14, 2,  2,
        38, 10, 2,  2,
        44,  0, 2, 16,
    };

    /** Creates new form KeyboardPanel. */
    public KeyboardPanel()
    {
        initComponents();
        setLayout(layout);

        // Create the buttons that correspond to keys. The numbers displayed
        // on the keycaps are internal key numbers -- similar, but not
        // identical, to scancode set 3.
        for (int key = 0; key < (keyPos.length / 4); key++)
        {
            buttons[key] = new JToggleButton(Integer.toHexString(key));

            GridBagConstraints constraints = new GridBagConstraints();

            constraints.gridx      = keyPos[key * 4];
            constraints.gridy      = keyPos[key * 4 + 1];
            constraints.gridwidth  = keyPos[key * 4 + 2];
            constraints.gridheight = keyPos[key * 4 + 3];
            constraints.fill       = GridBagConstraints.HORIZONTAL;

            buttons[key].setPreferredSize(null);
            this.add(buttons[key], constraints);
        }
        // Key 0x50 is always down on a real IBM keyboard. Don't disable it;
        // the user might want to test the error code that happens if this key
        // is ever up.
        buttons[0x50].setSelected(true);

        // Create the keyboard ID jumpers.
        for (int key = 0; key < 12; key++)
        {
            keyid[key] = new JCheckBox();

            GridBagConstraints constraints = new GridBagConstraints();

            constraints.gridx      = 40 + key + ((key > 5) ? 2 : 0);
            constraints.gridy      = 1;
            constraints.gridwidth  = 1;
            constraints.gridheight = 1;

            this.add(keyid[key], constraints);
        }

        // Create the LEDs.
        for (int l = 0; l < 3; l++)
        {
            led[l] = new JCheckBox(ledName[l]);

            GridBagConstraints constraints = new GridBagConstraints();

            constraints.gridx      = 40 + 4 * l;
            constraints.gridy      = 3;
            constraints.gridwidth  = 3;
            constraints.gridheight = 1;

            led[l].setEnabled(false);
            this.add(led[l], constraints);

        }
        // Create two captions for the two keyboard ID groups.
        for (int lbl = 0; lbl < 2; lbl++)
        {
            JLabel label = new JLabel((lbl == 0) ? "KBDID A" : "KBDID B");

            GridBagConstraints constraints = new GridBagConstraints();

            constraints.gridx      = 40 + 8 * lbl;
            constraints.gridy      = 0;
            constraints.gridwidth  = 6;
            constraints.gridheight = 1;

            this.add(label, constraints);
        }
        // Create the spacer panels.
        for (int spacer = 0; spacer < (spacerPos.length / 4); spacer++)
        {
            GridBagConstraints constraints = new GridBagConstraints();

            constraints.gridx      = spacerPos[spacer * 4];
            constraints.gridy      = spacerPos[spacer * 4 + 1];
            constraints.gridwidth  = spacerPos[spacer * 4 + 2];
            constraints.gridheight = spacerPos[spacer * 4 + 3];
            constraints.fill       = GridBagConstraints.BOTH;

            JPanel p = new JPanel();

            this.add(p, constraints);
        }
        // Create a row of dummy labels at the bottom. This is to ensure
        // every column has something in it, and so is not squeezed away
        // by the GridBagLayout.
        for (int col = 0; col < 54; col++)
        {
            GridBagConstraints constraints = new GridBagConstraints();

            constraints.gridx      = col;
            constraints.gridy      = 18;
            constraints.gridwidth  = 1;
            constraints.gridheight = 1;
            constraints.fill       = GridBagConstraints.BOTH;

            JLabel lbl = new JLabel("_^_");
            lbl.setForeground(getBackground());

            this.add(lbl, constraints);

        }

        setVisible(true);
    }

    // Read the identity jumpers and generate the keyboard ID therefrom.
    private void readID()
    {
        byte mask = 0x20;
        int n;
        byte result = -1;

        for (n = 0; n < 12; n++)
        {
            if (keyid[n].isSelected())
            {
                result &= ~mask;
            }
            mask = (byte)(mask >> 1);
            if (n == 5)  { kbID[0] = result; mask = 0x20; result = -1; }
            if (n == 11) { kbID[1] = result; mask = 0x20; result = -1; }
        }
    }


    /** Read the state of the keyboard jumpers (using a cross-thread call).
     *
     * @param set true to read set B, false to read set A.
     * @return the jumper settings, 0x00 - 0x3F.
     */
    public byte getID(boolean set)
    {
        Runnable r = new Runnable()
        {
            public void run()
            {
                readID();
            }
        };
        try
        {
            SwingUtilities.invokeAndWait(r);
        }
        catch (Exception ex)
        {
            return -1;
        }
        return kbID[set ? 1 : 0];
    }




    /** Get the pressed/released state of a key.
     *
     * @param key The key number (internal key number, 0-7F)
     * @return true if the key is down, false if it is up.
     */
    public boolean isKeyDown(final int key)
    {
        final boolean[] result = new boolean[1];

        Runnable r = new Runnable()
        {
            public void run()
            {
                result[0] = buttons[key & 0x7F].isSelected();
            }
        };
        try
        {
            SwingUtilities.invokeAndWait(r);
            return result[0];
        }
        catch (Exception ex)
        {
            return false;
        }
    }


    /** Set the state of the LEDs.
     *
     * @param b A bitmask of LEDs to be set. Only the bottom 3 bits matter.
     */
    public void setLEDs(byte b)
    {
//        System.out.printf("Set LEDs to %x\n", b);
        if (ledStatus == b)
        {
            return;
        }
        ledStatus = b;
        Runnable r = new Runnable()
        {
            public void run()
            {
                int b = ledStatus;

                for (int n = 0; n < 3; n++)
                {
                    led[n].setSelected((b & 1) != 0);
                    b = (byte)(b >> 1);
                }
            }
        };
        SwingUtilities.invokeLater(r);
    }



    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        setName("Form"); // NOI18N

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 400, Short.MAX_VALUE)
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 300, Short.MAX_VALUE)
        );
    }// </editor-fold>//GEN-END:initComponents


    // Variables declaration - do not modify//GEN-BEGIN:variables
    // End of variables declaration//GEN-END:variables

}
