/* KeyboardHost: Copyright 2011 John Elliott <jce@seasip.demon.co.uk>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package info.seasip.keyboardhost;

import java.awt.Component;
import java.io.*;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;

/** Intel 8048/8049/8050 microcontroller emulator.
 *
 * This does not completely emulate the 8048; there is no support for
 * interrupts (timer or otherwise), and instructions not used by the
 * IBM keyboard firmware ROM (such as "DA A") are untested.
 *
 * The emulation is run in a background thread, throttled so it runs at a
 * speed reasonably close to the original IBM keyboard.
 *
 * @author John Elliott
 */
public class MCU8048 extends Thread
{
    private byte a;
    private byte t, tScale;
    private byte psw;
    private byte rom[] = new byte[4096];
    private byte ram[] = new byte[1024];
    private short pc;
    private boolean tcnti = false;
    private boolean interrupt = false;
    private boolean mb = false;
    private boolean f1 = false;
    private boolean trace = false;
    private boolean timerFlag = false;
    private boolean timerTimes = false;
    private boolean timerCounts = false;
    private Thread workThread = null;
    private boolean stopped = false;
    private boolean running = false;
    private MCURunningListener listener = null;

    // Program status word bits
    private static final int CARRY     = 0x80;    // Bit 7 = carry
    private static final int HALFCARRY = 0x40;    // Bit 6 = half-carry
    private static final int FLAG0     = 0x20;    // Bit 5 = Flag 0
    private static final int REGBANK   = 0x10;    // Bit 4 = register bank

    private IOPort io = new DefaultIOPort();

    private Component parent;   // Used to display messages

    /** Construct a new, empty 8048 microcontroller.
     *
     * @param p Parent window (for message boxes).
     */
    public MCU8048(Component p)
    {
        int n;

        parent = p;
        for (n = 0; n < rom.length; n++) rom[n] = -1;
    }


    /** Is the emulation running?
     *
     * @return true if the emulation thread is running, else false.
     */
    public boolean isRunning()
    {
        return running;
    }


    /** Stop the emulation.
     *
     * @param wait If true, wait until emulation has stopped. Otherwise
     *            return before the emulation has stopped.
     */
    public void stopMCU(boolean wait)
    {
        if (running)
        {
            stopped = true;
            if (wait) while (running)
            {
                Thread.yield();
            }
        }
    }


    /** Reset the microcontroller.
     *
     * Stops the emulation (if running).
     * Resets all registers and RAM to default values.
     */
    public void reset()
    {
        stopMCU(true);
        
        a = 0;
        t = 0;
        tScale = 0;
        psw = 0;
        pc = 0;
        tcnti = false;
        interrupt = false;
        mb = false;
        f1 = false;
        trace = false;
        timerFlag = false;
        timerTimes = false;
        timerCounts = false;
        for (int n = 0; n < ram.length; n++)
        {
            ram[n] = -1;
        }
    }


    /** Start the emulation, using the passed ROM.
     *
     * @param r The ROM to use, which should be no more than 4k in size.
     */
    public void launch(byte [] r)
    {
        reset();
        System.arraycopy(r, 0, rom, 0, r.length);
        workThread = new Thread(this);
        workThread.setPriority(Thread.MIN_PRIORITY);
        workThread.start();

    }

    // Getters and setters
    public MCURunningListener getListener()
    {
        return listener;
    }

    public void setListener(MCURunningListener value)
    {
        listener = value;
    }

    public IOPort getIO()
    {
        return io;
    }

    public void setIO(IOPort value)
    {
        io = value;
    }


    // Accessors for the flag bits
    private boolean getCarry()
    {
        return (psw & CARRY) != 0;
    }

    private boolean getHalfCarry()
    {
        return (psw & HALFCARRY) != 0;
    }

    private boolean getFlag0()
    {
        return (psw & FLAG0) != 0;
    }

    private boolean getRB()
    {
        return (psw & REGBANK) != 0;
    }


    private void setCarry(boolean b)
    {
        if (b) psw |= CARRY;
        else   psw &= ~CARRY;
    }

    private void setHalfCarry(boolean b)
    {
        if (b) psw |= HALFCARRY;
        else   psw &= ~HALFCARRY;
    }

    private void setFlag0(boolean b)
    {
        if (b) psw |= FLAG0;
        else   psw &= ~FLAG0;
    }



    private void setRB(boolean b)
    {
        if (b) psw |= REGBANK;
        else   psw &= ~REGBANK;
    }


    // Add two 8-bit values (possibly plus a carry) and set the carry flag as
    // appropriate.
    private byte addc(byte arg1, byte arg2, boolean carry)
    {
        int halfres = (arg1 & 0x0F) + (arg2 & 0x0F);
        int res = (arg1 & 0xFF) + (arg2 & 0xFF);
        if (carry) 
        {
            ++res;
            ++halfres;
        }
        setHalfCarry((halfres & 0x10) != 0);
        setCarry((res & 0x100) != 0);
        return (byte)(res & 0xFF);
    }


    // Decimal adjust A. Completely untested.
    private void daa()
    {
        if (((a & 0x0F) > 9) || getHalfCarry())
        {
            a += 6;
            setCarry(true);
        }
        if (((a & 0xF0) > 0x90) || getCarry())
        {
            a += 96;
            setCarry(true);
        }
    }

    // Fetch the next byte from the ROM at PC, and increment PC.
    private byte fetch()
    {
        byte value = rom[pc];

        pc = (short)(((pc + 1) & 0x7FF) | (pc & 0x800));

        tick();
        return value;
    }


    private boolean lastT1 = false;

    // Emulate the timer. Only the timerTimes mode has had any testing at
    // all, and machine cycles are not emulated with any particular accuracy.
    private void tick()
    {
        if (timerTimes)     // If timer is counting machine cycles
        {
            ++tScale;       // Scale down so it increments every 32
            if (tScale == 32)
            {
                tScale = 0; // Wrap tScale at 32
                ++t;
                if (t == 0) // And when t rolls over, set the timer/counter flag
                {
                    timerFlag = true;
                }
            }
        }
        if (timerCounts)    // If we're counting low->high transitions on T1
        {
            boolean t1 = io.readT1();
            if (lastT1 != t1)   // T1 has changed
            {
                lastT1 = t1;
                if (t1)         // if it went high, bump T as above.
                {
                    ++t;
                    if (t == 0)
                    {
                        timerFlag = true;
                    }
                }
            }
        }
    }
    

    // Accessors for the 8048 registers R0-R7
    private byte getReg(int reg)
    {
        if (getRB())
        {
            reg = (reg & 7) + 0x18;
        }
        else
        {
            reg = (reg & 7);
        }
        return ram[reg];
    }

    private void setReg(int reg, byte value)
    {
        if (getRB())
        {
            reg = (reg & 7) + 0x18;
        }
        else
        {
            reg = (reg & 7);
        }
        store((byte)reg, value);
    }


    // 8048 return
    private void ret(boolean restoreFlags)
    {
        psw = (byte)((psw & 0xF8) | ((psw - 1) & 7));
        byte ret_h, ret_l;
        int sp = 8 + ((psw & 7) * 2);

        ret_h = ram[sp  ];
        ret_l = ram[sp+1];

        pc = (short)((ret_l & 0xFF) | ((ret_h << 8) & 0x0F00));
        if (restoreFlags)
        {
            psw = (byte)((psw & 0x0F) | (ret_h & 0xF0));
        }
    }


    // 8048 call
    private void call(int addr)
    {
        // Push return address
        byte ret_h = (byte)(((pc >> 8) & 0x0F) | (psw & 0xF0));
        byte ret_l = (byte)(pc & 0xFF);
        int sp = 8 + ((psw & 7) * 2);

        store((byte)sp,     ret_h);
        store((byte)(sp+1), ret_l);

        psw = (byte)((psw & 0xF8) | ((psw + 1) & 7));

        jump(addr);
    }


    // 8048 jump (possibly across pages)
    private void jump(int addr)
    {
        addr &= 0x7FF;
        if (mb) addr |= 0x800;
        pc = (short)addr;
    }

    // 8048 jump (within a page)
    private void nearjump(int addr)
    {
        addr &= 0xFF;
        pc = (short)((pc & 0xF00) | addr);
    }


    // Trace system state.
    private void printTrace(PrintStream ps, byte op)
    {
        int n;

        ps.printf("PC=%04x op=%02x A=%02x T=%02x ", pc-1, op & 0xFF, a, t);
        for (n = 0; n < 8; n++)
        {
            ps.printf("R%d=%02x ", n, getReg(n));
        }
        ps.printf("Cy=%d TF=%d RB=%d ", getCarry() ? 1 : 0, timerFlag ? 1 : 0, getRB() ? 1 : 0);
        ps.println("");
    }

    // Store a byte in controller RAM.
    private void store(byte addr, byte value)
    {
        //if (addr >= 0x20)
        //{
        //    System.out.printf("Write %02x to 0x%02x at PC=%04x\n", value, addr, pc);
        //}
        ram[addr & 0xFF] = value;
    }

    // Read a byte from controller RAM.
    private byte peek(byte addr)
    {
        return ram[addr & 0xFF];
    }


    /** Controller emulation main loop. Runs until the controller is
     * stopped by an outside influence, or an unrecognised instruction
     * is encountered.
     */
    @Override
    public void run()
    {
        pc = 0;
        tScale = 0;
        lastT1 = false;
        running = true;
        stopped = false;

        if (listener != null)   // If someone wants to be notified when we
        {                       // start running, notify them (using a cross-
            Runnable r = new Runnable() // -thread call).
            {
                public void run()
                {
                    listener.onMCUStartStop(true);
                }
            };
            SwingUtilities.invokeLater(r);
        }
        try
        {
            while (!stopped)    // Main emulation loop: Run for 500
            {                   // instructions, then sleep for a millisecond.
                for (int n = 0; n < 500; n++)
                {
                    step();
                    if (stopped)
                    {
                        break;
                    }
                }
                if (stopped)
                {
                    break;
                }
                try
                {
                    Thread.sleep(1);
                }
                catch (InterruptedException ex)
                {
                }
            }
        }
        catch(MCUException ex)  // Report an invalid instruction. Again, this
        {                       // is done with a cross-thread call.
            final String msg = ex.getMessage();

            Runnable r = new Runnable() {
                public void run()
                {
                    JOptionPane.showMessageDialog(parent, msg, "MCU 8048 exception", JOptionPane.ERROR_MESSAGE);
                }
            };
            SwingUtilities.invokeLater(r);
        }
        if (listener != null)       // MCU has stopped, so notify the UI.
        {
            Runnable r = new Runnable()
            {
                public void run()
                {
                    listener.onMCUStartStop(false);
                }
            };
            SwingUtilities.invokeLater(r);
        }
        running = false;
    }

    // Execute a single instruction.
    private void step() throws MCUException
    {
        byte op, arg, arg2;

//        if (pc >= 0x714 && pc <= 0x715)
//        {
//            trace = true;
//        }

        op = fetch();

        if (trace)
        {
            printTrace(System.out, op);
        }

        switch(op & 0xFF)
        {
            case 0x00:          // NOP
                break;
                                // No 0x01
            case 0x02:          // OUTL BUS, A
                io.writeBus(a);
                break;
            case 0x03:          // ADD a, nn
                arg = fetch();
                a = addc(a, arg, false);
                break;
            case 0x04:          // Jump1
                arg = fetch();
                jump(arg & 0xFF);
                break;
            case 0x05:          // EN I
                interrupt = true;
                break;
                                // No 0x06
            case 0x07:          // DEC A
                a--;
                break;
            case 0x08:          // INS A, BUS
                a = io.readBus();
                break;
            case 0x09:          // INS A, Pn
            case 0x0A:
//            case 0x0B:        // There is no P3
            case 0x0C:
            case 0x0D:
            case 0x0E:
            case 0x0F:
                a = io.readPort(op & 7);
                break;
            case 0x10:          // INC @R0
            case 0x11:          // INC @R1
                arg = getReg(op & 1);
                store(arg, (byte)(peek(arg) + 1));
                break;
            case 0x12:          // JB0 nn
                arg = fetch();
                if ((a & 0x01) != 0)
                {
                    nearjump(arg);
                }
                break;
            case 0x13:          // ADDC a, nn
                arg = fetch();
                a = addc(a, arg, getCarry());
                break;
            case 0x14:          // Call0
                arg = fetch();
                call(arg & 0xFF);
                break;
            case 0x15:          // DIS I
                interrupt = false;
                break;
            case 0x16:          // JTF nn
                arg = fetch();
                if (timerFlag)
                {
                    timerFlag = false;
                    nearjump(arg);
                }
                break;
            case 0x17:          // INC A
                a++;
                break;
            case 0x18:          // INC R0
            case 0x19:          // INC R1
            case 0x1A:          // INC R2
            case 0x1B:          // INC R3
            case 0x1C:          // INC R4
            case 0x1D:          // INC R5
            case 0x1E:          // INC R6
            case 0x1F:          // INC R7
                arg = getReg(op & 7);
                arg++;
                setReg(op & 7, arg);
                break;
            case 0x20:          // XCH A,@R0
            case 0x21:          // XCH A,@R1
                arg = getReg(op & 1);
                arg2 = peek(arg);
                store(arg, a);
                a = arg2;
                break;
                                // No 0x22
            case 0x23:          // MOV A, nn
                a = fetch();
                break;
            case 0x24:          // Jump1
                arg = fetch();
                jump(0x100 + (arg & 0xFF));
                break;
            case 0x25:          // EN TCNTI
                tcnti = true;
                break;
            case 0x26:          // JNT0 nn
                arg = fetch();
                if (!io.readT0())
                {
                    nearjump(arg);
                }
                break;
            case 0x27:          // CLR A
                a = 0;
                break;
            case 0x28:          // XCH A, R0
            case 0x29:          // XCH A, R1
            case 0x2A:          // XCH A, R2
            case 0x2B:          // XCH A, R3
            case 0x2C:          // XCH A, R4
            case 0x2D:          // XCH A, R5
            case 0x2E:          // XCH A, R6
            case 0x2F:          // XCH A, R7
                arg = getReg(op & 7);
                setReg(op & 7, a);
                a = arg;
                break;
            case 0x30:          // XCHD A,@R0
            case 0x31:          // XCHD A,@R1
                arg = getReg(op & 1);
                arg2 = (byte)((peek(arg) & 0x0F) | (a & 0xF0));
                store(arg, (byte)((peek(arg) & 0xF0) | (a & 0x0F)));
                a = arg2;
                break;
            case 0x32:          // JB1 nn
                arg = fetch();
                if ((a & 0x02) != 0)
                {
                    nearjump(arg);
                }
                break;
                                // No 0x33
            case 0x34:          // Call1
                arg = fetch();
                call((arg & 0xFF) | 0x100);
                break;
            case 0x35:          // DIS TCNTI
                tcnti = false;
                break;
            case 0x36:          // JT0 nn
                arg = fetch();
                if (io.readT0())
                {
                    nearjump(arg);
                }
                break;
            case 0x37:          // CPL A
                a ^= 0xFF;
                break;
            case 0x39:          // OUTL P1, A
            case 0x3A:          // OUTL P2, A
            case 0x3C:          // MOVD P4, A
            case 0x3D:          // MOVD P5, A
            case 0x3E:          // MOVD P6, A
            case 0x3F:          // MOVD P7, A
                io.writePort(op & 7, a);
                break;
            case 0x40:          // ORL A, @R0
            case 0x41:          // ORL A, @R1
                arg = peek(getReg(op & 1));
                a |= arg;
                break;
            case 0x42:          // MOV A, T
                a = t;
                break;
            case 0x43:          // ORL A, nn
                arg = fetch();
                a |= arg;
                break;
            case 0x44:          // Jump2
                arg = fetch();
                jump(0x200 + (arg & 0xFF));
                break;
            case 0x45:          // START CNT
                timerTimes = false;
                timerCounts = true;
                break;
            case 0x46:          // JNT1 nn
                arg = fetch();
                if (!io.readT1())
                {
                    nearjump(arg);
                }
                break;
            case 0x47:          // SWAP A
                a = (byte)(((a >> 4) & 0x0F) | ((a << 4) & 0xF0));
                break;
            case 0x48:          // ORL A, R0
            case 0x49:          // ORL A, R1
            case 0x4A:          // ORL A, R2
            case 0x4B:          // ORL A, R3
            case 0x4C:          // ORL A, R4
            case 0x4D:          // ORL A, R5
            case 0x4E:          // ORL A, R6
            case 0x4F:          // ORL A, R7
                a |= getReg(op & 7);
                break;
            case 0x50:          // ANL A, @R0
            case 0x51:          // ANL A, @R1
                arg = peek(getReg(op & 1));
                a &= arg;
                break;
            case 0x52:          // JB2 nn
                arg = fetch();
                if ((a & 0x04) != 0)
                {
                    nearjump(arg);
                }
                break;
            case 0x53:          // ANL A, nn
                arg = fetch();
                a &= arg;
                break;
            case 0x54:          // Call2
                arg = fetch();
                call((arg & 0xFF) | 0x200);
                break;
            case 0x55:          // STRT T
                timerTimes = true;
                timerCounts = false;
                break;
            case 0x56:          // JT1 nn
                arg = fetch();
                if (io.readT1())
                {
                    nearjump(arg);
                }
                break;
            case 0x57:          // DA A
                daa();
                break;
            case 0x58:          // ANL A, R0
            case 0x59:          // ANL A, R1
            case 0x5A:          // ANL A, R2
            case 0x5B:          // ANL A, R3
            case 0x5C:          // ANL A, R4
            case 0x5D:          // ANL A, R5
            case 0x5E:          // ANL A, R6
            case 0x5F:          // ANL A, R7
                a &= getReg(op & 7);
                break;
            case 0x60:          // ADD A, @R0
            case 0x61:          // ADD A, @R1
                arg = peek(getReg(op & 1));
                a = addc(a, arg, false);
                break;
            case 0x62:          // MOV T, A
                t = a;
                break;
                                // No 0x63
            case 0x64:          // Jump3
                arg = fetch();
                jump(0x300 + (arg & 0xFF));
                break;
            case 0x65:          // STOP TCNT
                timerTimes = false;
                timerCounts = false;
                break;
                                // No 0x66, should be JNF1
            case 0x67:          // RRC A
                arg = (byte)((a >> 1) & 0x7F);
                if (getCarry()) arg |= 0x80;
                setCarry( (a & 1) != 0);
                a = arg;
                break;

            case 0x68:          // ADD A, R0
            case 0x69:          // ADD A, R1
            case 0x6A:          // ADD A, R2
            case 0x6B:          // ADD A, R3
            case 0x6C:          // ADD A, R4
            case 0x6D:          // ADD A, R5
            case 0x6E:          // ADD A, R6
            case 0x6F:          // ADD A, R7
                arg = getReg(op & 7);
                a = addc(a, arg, false);
                break;
            case 0x70:          // ADD A, @R0
            case 0x71:          // ADD A, @R1
                arg = peek(getReg(op & 1));
                a = addc(a, arg, getCarry());
                break;
             case 0x72:          // JB3 nn
                arg = fetch();
                if ((a & 0x08) != 0)
                {
                    nearjump(arg);
                }
                break;
                                // No 0x73
            case 0x74:          // Call3
                arg = fetch();
                call((arg & 0xFF) | 0x300);
                break;
//            case 0x75:          // TODO: ENT0 CLK, unimplemented
            case 0x76:          // JF1
                arg = fetch();
                if (f1)
                {
                    nearjump(arg);
                }
                break;
            case 0x77:          // RR A
                a = (byte)(((a >> 1) & 0x7F) | ((a << 7) & 0x80));
                break;
            case 0x78:          // ADDC A, R0
            case 0x79:          // ADDC A, R1
            case 0x7A:          // ADDC A, R2
            case 0x7B:          // ADDC A, R3
            case 0x7C:          // ADDC A, R4
            case 0x7D:          // ADDC A, R5
            case 0x7E:          // ADDC A, R6
            case 0x7F:          // ADDC A, R7
                arg = getReg(op & 7);
                a = addc(a, arg, getCarry());
                break;
            case 0x80:          // MOVX A, @R0
            case 0x81:          // MOVX A, @R1
                arg = getReg(op & 1);
                a = io.readExtMem(arg);
                break;
                                // No 0x82
            case 0x83:          // RET
                ret(false);
                break;
            case 0x84:          // Jump4
                arg = fetch();
                jump(0x400 + (arg & 0xFF));
                break;
            case 0x85:          // CLR F0
                setFlag0(false);
                break;
            case 0x86:          // JNI nn
                arg = fetch();
                if (io.readI() == false)
                {
                    nearjump(arg);
                }
                break;
                                // No 0x87
            case 0x88:
                arg = fetch();  // ORL Bus, nn
                arg |= io.readBus();
                io.writeBus(arg);
                break;
            case 0x89:          // ORL P1, nn
            case 0x8A:          // ORL P2, nn
            case 0x8C:          // ORLD P4, nn
            case 0x8D:          // ORLD P5, nn
            case 0x8E:          // ORLD P6, nn
            case 0x8F:          // ORLD P7, nn
                arg = fetch();
                arg |= io.readPort(op & 7);
                io.writePort(op & 7, arg);
                break;
            case 0x90:          // MOVX A, @R0
            case 0x91:          // MOVX A, @R1
                arg = getReg(op & 1);
                io.writeExtMem(arg, a);
                break;

            case 0x92:          // JB4 nn
                arg = fetch();
                if ((a & 0x10) != 0)
                {
                    nearjump(arg);
                }
                break;
            case 0x93:          // RETR
                ret(true);
                break;
            case 0x94:          // Call4
                arg = fetch();
                call((arg & 0xFF) | 0x400);
                break;
            case 0x95:          // CPL F0
                setFlag0(!getFlag0());
                break;
            case 0x96:          // JNZ nn
                arg = fetch();
                if (a != 0)
                {
                    nearjump(arg);
                }
                break;
            case 0x97:          // CLR C
                psw &= 0x7F;
                break;
            case 0x98:
                arg = fetch();  // ANLD Bus, nn
                arg &= io.readBus();
                io.writeBus(arg);
                break;
            case 0x99:          // ANL P1, nn
            case 0x9A:          // ANL P2, nn
            case 0x9C:          // ANLD P4, nn
            case 0x9D:          // ANLD P5, nn
            case 0x9E:          // ANLD P6, nn
            case 0x9F:          // ANLD P7, nn
                arg = fetch();
                arg &= io.readPort(op & 7);
                io.writePort(op & 7, arg);
                break;

            case 0xA0:          // MOV @R0, A
            case 0xA1:          // MOV @R1, A
                arg = getReg(op & 1);
                store(arg, a);
                break;
                                // No 0xA2
            case 0xA3:          // MOVP A, @A
                a = rom[(pc & 0xF00) | (a & 0xFF)];
                break;
            case 0xA4:          // Jump5
                arg = fetch();
                jump(0x500 + (arg & 0xFF));
                break;
            case 0xA5:          // CLR F1
                f1 = false;
                break;
            case 0xA7:          // CPL C
                //setCarry(!getCarry());
                psw ^= CARRY;
                break;
            case 0xA8:          // MOV R0, A
            case 0xA9:          // MOV R1, A
            case 0xAA:          // MOV R2, A
            case 0xAB:          // MOV R3, A
            case 0xAC:          // MOV R4, A
            case 0xAD:          // MOV R5, A
            case 0xAE:          // MOV R6, A
            case 0xAF:          // MOV R7, A
                setReg(op & 7, a);
                break;

            case 0xB0:          // MOV @R0, nn
            case 0xB1:          // MOV @R1, nn
                arg = getReg(op & 1);
                arg2 = fetch();
                store(arg, arg2);
                break;
            case 0xB2:          // JB5 nn
                arg = fetch();
                if ((a & 0x20) != 0)
                {
                    nearjump(arg);
                }
                break;
            case 0xB3:          // JMPP @A
                arg = rom[(pc & 0x0F00) | (a & 0xFF)];
                nearjump(arg);
                break;

            case 0xB4:          // Call5
                arg = fetch();
                call((arg & 0xFF) | 0x500);
                break;
            case 0xB5:          // CPL F1
                f1 = !f1;
                break;
            case 0xB6:          // JF0 nn
                arg = fetch();
                if (getFlag0())
                {
                    nearjump(arg);
                }
                break;
                                // No 0xB7
            case 0xB8:          // MOV R0, nn
            case 0xB9:          // MOV R1, nn
            case 0xBA:          // MOV R2, nn
            case 0xBB:          // MOV R3, nn
            case 0xBC:          // MOV R4, nn
            case 0xBD:          // MOV R5, nn
            case 0xBE:          // MOV R6, nn
            case 0xBF:          // MOV R7, nn
                arg = fetch();
                setReg(op & 7, arg);
                break;
                                // No 0xC0-0xC3
            case 0xC4:          // Jump6
                arg = fetch();
                jump(0x600 + (arg & 0xFF));
                break;
            case 0xC5:          // SEL RB0
                setRB(false);
                break;
            case 0xC6:          // JZ nn
                arg = fetch();
                if (a == 0)
                {
                    nearjump(arg);
                }
                break;
            case 0xC7:          // MOV A, psw
                a = psw;
                break;
            case 0xC8:          // DEC R0
            case 0xC9:          // DEC R1
            case 0xCA:          // DEC R2
            case 0xCB:          // DEC R3
            case 0xCC:          // DEC R4
            case 0xCD:          // DEC R5
            case 0xCE:          // DEC R6
            case 0xCF:          // DEEC R7
                arg = getReg(op & 7);
                arg--;
                setReg(op & 7, arg);
                break;
            case 0xD0:          // XRL A, @R0
            case 0xD1:          // XRL A, @R1
                arg = getReg(op & 1);
                a ^= peek(arg);
                break;
            case 0xD2:          // JB6 nn
                arg = fetch();
                if ((a & 0x40) != 0)
                {
                    nearjump(arg);
                }
                break;
            case 0xD3:          // XRL A, nn
                arg = fetch();
                a ^= arg;
                break;
            case 0xD4:          // Call6
                arg = fetch();
                call((arg & 0xFF) | 0x600);
                break;
            case 0xD5:          // SEL RB0
                setRB(true);
                break;
            // There is no D6, but it ought to be JNZ
            case 0xD7:          // MOV PSW, A
                psw = a;
                break;
            case 0xD8:          // XRL A, R0
            case 0xD9:          // XRL A, R1
            case 0xDA:          // XRL A, R2
            case 0xDB:          // XRL A, R3
            case 0xDC:          // XRL A, R4
            case 0xDD:          // XRL A, R5
            case 0xDE:          // XRL A, R6
            case 0xDF:          // XRL A, R7
                a ^= getReg(op & 7);
                break;
            // E0-E1 would be DJNZ @R0, DJNZ @R1
            case 0xE3:          // MOVP3 A, @A
                a = rom[0x300 | (a & 0xFF)];
                break;
            case 0xE4:          // Jump7
                arg = fetch();
                jump(0x700 + (arg & 0xFF));
                break;
            case 0xE5:          // SEL MB0
                mb = false;
                break;
            case 0xE6:          // JNC nn
                arg = fetch();
                if (!getCarry())
                {
                    nearjump(arg);
                }
                break;
            case 0xE7:          // RL A
                a = (byte)(((a << 1) & 0xFE) | ((a >> 7) & 1));
                break;
            case 0xE8:          // DJNZ R0, nn
            case 0xE9:          // DJNZ R1, nn
            case 0xEA:          // DJNZ R2, nn
            case 0xEB:          // DJNZ R3, nn
            case 0xEC:          // DJNZ R4, nn
            case 0xED:          // DJNZ R5, nn
            case 0xEE:          // DJNZ R6, nn
            case 0xEF:          // DJNZ R7, nn
                arg = getReg(op & 7);
                arg--;
                setReg(op & 7, arg);
                arg2 = fetch();
                if (arg != 0)
                {
                    nearjump(arg2);
                }
                break;

            case 0xF0:          // MOV A, @R0
            case 0xF1:          // MOV A, @R1
                arg = getReg(op & 1);
                a = peek(arg);
                break;
            case 0xF2:          // JB7 nn
                arg = fetch();
                if ((a & 0x80) != 0)
                {
                    nearjump(arg);
                }
                break;
            case 0xF4:          // Call7
                arg = fetch();
                call((arg & 0xFF) | 0x700);
                break;
            case 0xF5:          // SEL MB1
                mb = true;
                break;
            case 0xF6:          // JC nn
                arg = fetch();
                if (getCarry())
                {
                    nearjump(arg);
                }
                break;
            case 0xF7:          // RLC A
                arg = (byte)((a << 1) & 0xFE);
                if (getCarry()) arg |= 0x01;
                setCarry((a & 0x80) != 0);
                a = arg;
                break;

            case 0xF8:          // MOV A, R0
            case 0xF9:          // MOV A, R1
            case 0xFA:          // MOV A, R2
            case 0xFB:          // MOV A, R3
            case 0xFC:          // MOV A, R4
            case 0xFD:          // MOV A, R5
            case 0xFE:          // MOV A, R6
            case 0xFF:          // MOV A, R7
                a = getReg(op & 7);
                break;

            default:
                throw new MCUException(String.format("Unknown 8048 opcode 0x%02x at PC=%04x",
                                        op & 0xFF, pc & 0xFFF));
        }
    }
}
