/* KeyboardHost: Copyright 2011 John Elliott <jce@seasip.demon.co.uk>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


package info.seasip.keyboardhost;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

/** RomPanel provides overall control of KeyboardHost. It allows ROM images
 * to be loaded, saved, customized and executed.
 *
 * @author John Elliott
 */
public class RomPanel extends javax.swing.JPanel implements MCURunningListener
{
    private byte [] rom5170 = null; // IBM's ROM image from the AT keyboard
    private byte [] rom5271 = null; // IBM's ROM image from the 3270PC keyboard
    private byte [] rom1k   = null; // 1k composite ROM image
    private byte [] rom2k   = null; // 2k composite ROM image

    // Magic number for the AT keyboard ROM: its part number, 1503099.
    private final byte[] magic5170 = new byte[] { '1', '5', '0', '3', '0', '9', '9' };

    // Magic number for the 3270PC keyboard ROM: the bytes that happen to be
    // where the part number is on the AT ROM.
    private final byte[] magic5271 = new byte[] { 0x03, 0x17, 0x28, (byte)0xE8, (byte)0xCC, 0x53, 0x03 };

    // References to the rest of the UI: the keyboard and host panels, and
    // the microcontroller emulator.
    private KeyboardPanel keyboardPanel = null;
    private HostPanel hostPanel = null;
    private MCU8048 mcu = null;

    /** Creates new form RomPanel */
    public RomPanel()
    {
        initComponents();

        // Attempt to load the IBM ROMs from the current directory, failing
        // silently if there is an error.
        load5170(true);
        load5271(true);
    }

    // Setters and getters for the references.
    public void setMCU(MCU8048 value)
    {
        mcu = value;
        mcu.setListener(this);
        enableControls();
    }


    public MCU8048 getMCU()
    {
        return mcu;
    }
    

    public void setKeyboardPanel(KeyboardPanel kp)
    {
        keyboardPanel = kp;
    }

    public KeyboardPanel getKeyboardPanel()
    {
        return keyboardPanel;
    }

    public void setHostPanel(HostPanel hp)
    {
        hostPanel = hp;
    }

    public HostPanel getHostPanel()
    {
        return hostPanel;
    }



    // Load a ROM image. We support up to 4k of ROM (which is what an 8050
    // would hold). Return an array of bytes giving the exact size.
    private byte [] loadRom(String filename) throws IOException, FileNotFoundException
    {
        byte [] buffer = new byte[4096];

        FileInputStream is = new FileInputStream(filename);
        int len = is.read(buffer);
        is.close();
        if (len < 4096)
        {
            byte [] buf2 = new byte[len];
            System.arraycopy(buffer, 0, buf2, 0, len);
            return buf2;
        }

        return buffer;
    }

    // Save a ROM image.
    private void saveRom(File filename, byte [] rom)
    {
        try
        {
            FileOutputStream os = new FileOutputStream(filename);
            os.write(rom);
            os.close();
        }
        catch(IOException ex)
        {
            JOptionPane.showMessageDialog(this, ex.getMessage(), "Save ROM", JOptionPane.ERROR_MESSAGE);            
        }
    }


    // When the microcontroller starts or stops, enable/disable the buttons.
    public void onMCUStartStop(boolean running)
    {
        enableControls();
    }


    // Enable / disable the various buttons:
    private void enableControls()
    {
        // * Hybrid ROMs can only be created if we have both IBM ROM images
        //  loaded.
        createHybrid1K.setEnabled(rom5170 != null && rom5271 != null);
        saveHybrid1K.setEnabled(rom1k != null);
        createHybrid2K.setEnabled(rom5170 != null && rom5271 != null);
        saveHybrid2K.setEnabled(rom2k != null);

        // * The ROMs can only be customised if they are loaded.
        custom1k.setEnabled(rom1k != null && rom1k.length >= 1024);
        custom2k.setEnabled(rom2k != null && rom2k.length >= 2048);

        // * The MCU can only be started if it is not running and has a
        //  valid ROM.
        launch5170.setEnabled(rom5170 != null && mcu != null && !mcu.isRunning());
        launch5271.setEnabled(rom5271 != null && mcu != null && !mcu.isRunning());
        launch1k.setEnabled(rom1k != null && mcu != null && !mcu.isRunning());
        launch2k.setEnabled(rom2k != null && mcu != null && !mcu.isRunning());

        // * The MCU can only be stopped if it is running.
        stopMCU.setEnabled(mcu != null && mcu.isRunning());

        // Update MCU status.
        if (mcu == null)
        {
            mcuStatus.setText("MCU not present.");
        }
        else if (mcu.isRunning())
        {
            mcuStatus.setText("MCU running.");
        }
        else mcuStatus.setText("MCU stopped.");

    }


    // Load the IBM AT keyboard ROM. If initial is true, fail without displaying
    // an alert; otherwise display one.
    private void load5170(boolean initial)
    {
        try
        {
            load5170();
            label5170.setText("IBM AT ROM image: Loaded " + name5170.getText());
        }
        catch (Exception ex)
        {
            if (!initial)
            {
                JOptionPane.showMessageDialog(this, ex.getMessage(), "Load ROM", JOptionPane.ERROR_MESSAGE);
            }
            else
            {
                label5170.setText("IBM 5170 ROM image:  " + ex.getMessage());
                rom5170 = null;
            }
        }
        enableControls();
    }


    // Load the IBM 3270PC keyboard ROM. If initial is true, fail without
    // displaying an alert; otherwise display one.
    private void load5271(boolean initial)
    {
        try
        {
            load5271();
            label5271.setText("IBM 3270PC ROM image: Loaded " + name5271.getText());
        }
        catch (Exception ex)
        {
            if (!initial)
            {
                JOptionPane.showMessageDialog(this, ex.getMessage(), "Load ROM", JOptionPane.ERROR_MESSAGE);
            }
            else
            {
                label5271.setText("IBM 3270PC ROM image:  " + ex.getMessage());
                rom5271 = null;
            }
        }
        enableControls();
    }

    // Load a ROM image into the 'custom 1k' slot -- though in fact, it can
    // be any size up to 4k.
    private void load1k(boolean initial)
    {
        try
        {
            load1k();
            labelHybrid1K.setText("Hybrid 1k ROM image: Loaded " + name1k.getText());
        }
        catch (Exception ex)
        {
            if (!initial)
            {
                JOptionPane.showMessageDialog(this, ex.getMessage(), "Load ROM", JOptionPane.ERROR_MESSAGE);
            }
            else
            {
                labelHybrid1K.setText("Hybrid 1k ROM image:  " + ex.getMessage());
                rom1k = null;
            }
        }
        enableControls();
    }


    // Load a ROM image into the 'custom 2k' slot -- though in fact, it can
    // be any size up to 4k.
    private void load2k(boolean initial)
    {
        try
        {
            load2k();
            labelHybrid2K.setText("Hybrid 2k ROM image: Loaded " + name2k.getText());
        }
        catch (Exception ex)
        {
            if (!initial)
            {
                JOptionPane.showMessageDialog(this, ex.getMessage(), "Load ROM", JOptionPane.ERROR_MESSAGE);
            }
            else
            {
                labelHybrid2K.setText("Hybrid 2k ROM image:  " + ex.getMessage());
                rom1k = null;
            }
        }
        enableControls();
    }




    // Load the IBM AT keyboard ROM. Throws an exception if the checksum is
    // bad or the magic number is not present.
    private void load5170() throws Exception
    {
        String filename = name5170.getText();
        try
        {
            byte [] result = loadRom(filename);
            if (!verifyChecksum(result))
            {
                throw new Exception(filename + ": Bad checksum");
            }
            if (!verifySignature(result, magic5170, 0x3C4))
            {
                throw new Exception(filename + ": Bad signature");
            }
            rom5170 = result;
        }
        catch (FileNotFoundException ex)
        {
            throw new Exception("Not found: " + ex.getMessage(), ex);
        }
        catch (IOException ex)
        {
            throw new Exception("Not loaded: " + ex.getMessage(), ex);
        }
        enableControls();
    }



    // Load the IBM 3270PC keyboard ROM. Throws an exception if the checksum is
    // bad or the magic number is not present.
    private void load5271() throws Exception
    {
        String filename = name5271.getText();
        try
        {
            byte [] result = loadRom(filename);
            if (!verifyChecksum(result))
            {
                throw new Exception(filename + ": Bad checksum");
            }
            if (!verifySignature(result, magic5271, 0x3C4))
            {
                throw new Exception(filename + ": Bad signature");
            }
            rom5271 = result;
        }
        catch (FileNotFoundException ex)
        {
            throw new Exception("Not found: " + ex.getMessage(), ex);
        }
        catch (IOException ex)
        {
            throw new Exception("Not loaded: " + ex.getMessage(), ex);
        }
        enableControls();
    }


    // Load into the 'composite 1k' slot. Warns if the checksum is bad, but
    // does not abort.
    private void load1k() throws Exception
    {
        String filename = name1k.getText();
        try
        {
            byte [] result = loadRom(filename);
            if (!verifyChecksum(result))
            {
                JOptionPane.showMessageDialog(this, filename + ": Checksum is not zero.", "Load ROM",
                        JOptionPane.WARNING_MESSAGE);
            }
            rom1k = result;
        }
        catch (FileNotFoundException ex)
        {
            throw new Exception("Not found: " + ex.getMessage(), ex);
        }
        catch (IOException ex)
        {
            throw new Exception("Not loaded: " + ex.getMessage(), ex);
        }
        enableControls();
    }

    // Load into the 'composite 2k' slot. Warns if the checksum is bad, but
   // does not abort.
   private void load2k() throws Exception
    {
        String filename = name2k.getText();
        try
        {
            byte [] result = loadRom(filename);
            if (!verifyChecksum(result))
            {
                JOptionPane.showMessageDialog(this, filename + ": Checksum is not zero.", "Load ROM",
                        JOptionPane.WARNING_MESSAGE);
            }
            rom2k = result;
        }
        catch (FileNotFoundException ex)
        {
            throw new Exception("Not found: " + ex.getMessage(), ex);
        }
        catch (IOException ex)
        {
            throw new Exception("Not loaded: " + ex.getMessage(), ex);
        }
        enableControls();
    }


    // Compare the signature of a ROM with a magic number.
    private boolean verifySignature(byte [] rom, byte [] signature, int base)
    {
        int n;
        
        for (n = 0; n < signature.length; n++)
        {
            if (rom[base + n] != signature[n])
            {
                return false;
            }
        }
        return true;
    }

    // Verify that the 8-bit checksum of all bytes in a ROM is 0.
    private boolean verifyChecksum(byte [] rom)
    {
        byte sum = 0;

        for (int n = 0; n < rom.length; n++)
        {
            sum += rom[n];
        }
        return (sum == 0);
    }


    
    // Create the 1k hybrid ROM image.
    private void doCreateHybrid1K()
    {
        if (rom5271 == null || rom5170 == null)
        {
            return;
        }
        rom1k = RomPatcher.createHybrid1k(rom5170, rom5271);

        labelHybrid1K.setText("Hybrid 1k ROM image: Created");
        enableControls();
    }


    // Create the 2k hybrid ROM image.
    private void doCreateHybrid2K()
    {
        int n;

        if (rom5271 == null || rom5170 == null)
        {
            return;
        }
        rom2k = RomPatcher.createHybrid2k(rom5170, rom5271);
        labelHybrid2K.setText("Hybrid 2k ROM image: Created");
        enableControls();
    }



    // Launch the MCU with a given ROM image.
    private void launch(byte [] rom)
    {
        if (rom == null || mcu == null || mcu.isRunning())
        {
            return;
        }
        if (keyboardPanel != null)
        {
            keyboardPanel.setVisible(true);
        }
        if (hostPanel != null)
        {
            hostPanel.setVisible(true);
            hostPanel.clearBuffer();
        }
        mcu.launch(rom);
    }


    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        label5170 = new javax.swing.JLabel();
        name5170 = new javax.swing.JTextField();
        browse5170 = new javax.swing.JButton();
        label5271 = new javax.swing.JLabel();
        name5271 = new javax.swing.JTextField();
        browse5271 = new javax.swing.JButton();
        labelHybrid1K = new javax.swing.JLabel();
        createHybrid1K = new javax.swing.JButton();
        saveHybrid1K = new javax.swing.JButton();
        launch5170 = new javax.swing.JButton();
        launch5271 = new javax.swing.JButton();
        launch1k = new javax.swing.JButton();
        mcuStatus = new javax.swing.JLabel();
        stopMCU = new javax.swing.JButton();
        labelHybrid2K = new javax.swing.JLabel();
        createHybrid2K = new javax.swing.JButton();
        saveHybrid2K = new javax.swing.JButton();
        launch2k = new javax.swing.JButton();
        browse1k = new javax.swing.JButton();
        browse2k = new javax.swing.JButton();
        name1k = new javax.swing.JTextField();
        name2k = new javax.swing.JTextField();
        custom2k = new javax.swing.JButton();
        custom1k = new javax.swing.JButton();

        setName("Form"); // NOI18N

        org.jdesktop.application.ResourceMap resourceMap = org.jdesktop.application.Application.getInstance(info.seasip.keyboardhost.KeyboardHostApp.class).getContext().getResourceMap(RomPanel.class);
        label5170.setText(resourceMap.getString("lbl5170.text")); // NOI18N
        label5170.setName("lbl5170"); // NOI18N

        name5170.setText(resourceMap.getString("filename5170.text")); // NOI18N
        name5170.setName("filename5170"); // NOI18N

        browse5170.setText(resourceMap.getString("browse5170.text")); // NOI18N
        browse5170.setName("browse5170"); // NOI18N
        browse5170.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                browse5170ActionPerformed(evt);
            }
        });

        label5271.setText(resourceMap.getString("lbl5271.text")); // NOI18N
        label5271.setName("lbl5271"); // NOI18N

        name5271.setText(resourceMap.getString("filename5271.text")); // NOI18N
        name5271.setName("filename5271"); // NOI18N

        browse5271.setText(resourceMap.getString("browse5271.text")); // NOI18N
        browse5271.setName("browse5271"); // NOI18N
        browse5271.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                browse5271ActionPerformed(evt);
            }
        });

        labelHybrid1K.setText(resourceMap.getString("labelHybrid1K.text")); // NOI18N
        labelHybrid1K.setName("labelHybrid1K"); // NOI18N

        createHybrid1K.setText(resourceMap.getString("createHybrid1K.text")); // NOI18N
        createHybrid1K.setName("createHybrid1K"); // NOI18N
        createHybrid1K.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                createHybrid1KActionPerformed(evt);
            }
        });

        saveHybrid1K.setText(resourceMap.getString("saveHybrid1K.text")); // NOI18N
        saveHybrid1K.setName("saveHybrid1K"); // NOI18N
        saveHybrid1K.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                saveHybrid1KActionPerformed(evt);
            }
        });

        launch5170.setText(resourceMap.getString("launch5170.text")); // NOI18N
        launch5170.setName("launch5170"); // NOI18N
        launch5170.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                launch5170ActionPerformed(evt);
            }
        });

        launch5271.setText(resourceMap.getString("launch5271.text")); // NOI18N
        launch5271.setName("launch5271"); // NOI18N
        launch5271.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                launch5271ActionPerformed(evt);
            }
        });

        launch1k.setText(resourceMap.getString("launch1k.text")); // NOI18N
        launch1k.setName("launch1k"); // NOI18N
        launch1k.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                launch1kActionPerformed(evt);
            }
        });

        mcuStatus.setText(resourceMap.getString("mcuStatus.text")); // NOI18N
        mcuStatus.setName("mcuStatus"); // NOI18N

        stopMCU.setText(resourceMap.getString("stopMCU.text")); // NOI18N
        stopMCU.setName("stopMCU"); // NOI18N
        stopMCU.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                stopMCUActionPerformed(evt);
            }
        });

        labelHybrid2K.setText(resourceMap.getString("labelHybrid2K.text")); // NOI18N
        labelHybrid2K.setName("labelHybrid2K"); // NOI18N

        createHybrid2K.setText(resourceMap.getString("createHybrid2K.text")); // NOI18N
        createHybrid2K.setName("createHybrid2K"); // NOI18N
        createHybrid2K.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                createHybrid2KActionPerformed(evt);
            }
        });

        saveHybrid2K.setText(resourceMap.getString("saveHybrid2K.text")); // NOI18N
        saveHybrid2K.setName("saveHybrid2K"); // NOI18N
        saveHybrid2K.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                saveHybrid2KActionPerformed(evt);
            }
        });

        launch2k.setText(resourceMap.getString("launch2k.text")); // NOI18N
        launch2k.setName("launch2k"); // NOI18N
        launch2k.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                launch2kActionPerformed(evt);
            }
        });

        browse1k.setText(resourceMap.getString("browse1k.text")); // NOI18N
        browse1k.setName("browse1k"); // NOI18N
        browse1k.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                browse1kActionPerformed(evt);
            }
        });

        browse2k.setText(resourceMap.getString("browse2k.text")); // NOI18N
        browse2k.setName("browse2k"); // NOI18N
        browse2k.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                browse2kActionPerformed(evt);
            }
        });

        name1k.setText(resourceMap.getString("name1k.text")); // NOI18N
        name1k.setName("name1k"); // NOI18N

        name2k.setText(resourceMap.getString("name2k.text")); // NOI18N
        name2k.setName("name2k"); // NOI18N

        custom2k.setText(resourceMap.getString("custom2k.text")); // NOI18N
        custom2k.setName("custom2k"); // NOI18N
        custom2k.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                custom2kActionPerformed(evt);
            }
        });

        custom1k.setText(resourceMap.getString("custom1k.text")); // NOI18N
        custom1k.setName("custom1k"); // NOI18N
        custom1k.addActionListener(new java.awt.event.ActionListener() {
            public void actionPerformed(java.awt.event.ActionEvent evt) {
                custom1kActionPerformed(evt);
            }
        });

        javax.swing.GroupLayout layout = new javax.swing.GroupLayout(this);
        this.setLayout(layout);
        layout.setHorizontalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addGroup(layout.createSequentialGroup()
                        .addContainerGap()
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                            .addGroup(layout.createSequentialGroup()
                                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                                    .addComponent(label5271, javax.swing.GroupLayout.DEFAULT_SIZE, 303, Short.MAX_VALUE)
                                    .addComponent(label5170, javax.swing.GroupLayout.DEFAULT_SIZE, 303, Short.MAX_VALUE))
                                .addGap(117, 117, 117))
                            .addGroup(layout.createSequentialGroup()
                                .addGap(21, 21, 21)
                                .addComponent(name5170, javax.swing.GroupLayout.DEFAULT_SIZE, 399, Short.MAX_VALUE))
                            .addGroup(layout.createSequentialGroup()
                                .addGap(18, 18, 18)
                                .addComponent(name5271, javax.swing.GroupLayout.DEFAULT_SIZE, 402, Short.MAX_VALUE))
                            .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, layout.createSequentialGroup()
                                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.TRAILING)
                                    .addGroup(javax.swing.GroupLayout.Alignment.LEADING, layout.createSequentialGroup()
                                        .addGap(18, 18, 18)
                                        .addComponent(createHybrid1K)
                                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                                        .addComponent(name1k, javax.swing.GroupLayout.DEFAULT_SIZE, 237, Short.MAX_VALUE))
                                    .addGroup(javax.swing.GroupLayout.Alignment.LEADING, layout.createSequentialGroup()
                                        .addGap(18, 18, 18)
                                        .addComponent(createHybrid2K)
                                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                                        .addComponent(name2k, javax.swing.GroupLayout.DEFAULT_SIZE, 237, Short.MAX_VALUE))
                                    .addComponent(labelHybrid2K, javax.swing.GroupLayout.Alignment.LEADING, javax.swing.GroupLayout.DEFAULT_SIZE, 316, Short.MAX_VALUE)
                                    .addComponent(labelHybrid1K, javax.swing.GroupLayout.Alignment.LEADING, javax.swing.GroupLayout.DEFAULT_SIZE, 316, Short.MAX_VALUE))
                                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                                    .addComponent(custom1k, javax.swing.GroupLayout.PREFERRED_SIZE, 92, javax.swing.GroupLayout.PREFERRED_SIZE)
                                    .addComponent(custom2k))))
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                            .addComponent(browse5170)
                            .addComponent(browse5271)
                            .addComponent(browse1k)
                            .addComponent(browse2k))
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                            .addComponent(saveHybrid1K)
                            .addComponent(saveHybrid2K))
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING, false)
                            .addComponent(launch5170, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                            .addComponent(launch5271, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                            .addComponent(launch1k, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                            .addComponent(launch2k, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                            .addComponent(stopMCU, javax.swing.GroupLayout.DEFAULT_SIZE, 85, Short.MAX_VALUE)))
                    .addGroup(layout.createSequentialGroup()
                        .addGap(12, 12, 12)
                        .addComponent(mcuStatus, javax.swing.GroupLayout.PREFERRED_SIZE, 410, javax.swing.GroupLayout.PREFERRED_SIZE)))
                .addContainerGap())
        );
        layout.setVerticalGroup(
            layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(layout.createSequentialGroup()
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addGroup(layout.createSequentialGroup()
                        .addGap(35, 35, 35)
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                            .addComponent(browse5170)
                            .addComponent(launch5170)))
                    .addGroup(layout.createSequentialGroup()
                        .addContainerGap()
                        .addComponent(label5170)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(name5170, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addComponent(label5271)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                            .addComponent(name5271, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                            .addComponent(browse5271)
                            .addComponent(launch5271))))
                .addGap(8, 8, 8)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                        .addComponent(browse2k)
                        .addComponent(saveHybrid2K)
                        .addComponent(launch2k)
                        .addComponent(custom2k))
                    .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, layout.createSequentialGroup()
                        .addComponent(labelHybrid1K)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                            .addComponent(createHybrid1K)
                            .addComponent(name1k, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                            .addComponent(browse1k)
                            .addComponent(saveHybrid1K)
                            .addComponent(launch1k)
                            .addComponent(custom1k, javax.swing.GroupLayout.PREFERRED_SIZE, 29, javax.swing.GroupLayout.PREFERRED_SIZE))
                        .addGap(18, 18, 18)
                        .addComponent(labelHybrid2K)
                        .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                        .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                            .addComponent(createHybrid2K)
                            .addComponent(name2k, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))))
                .addGap(18, 18, 18)
                .addGroup(layout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(stopMCU)
                    .addComponent(mcuStatus))
                .addContainerGap())
        );
    }// </editor-fold>//GEN-END:initComponents

    private void browse5170ActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_browse5170ActionPerformed
        JFileChooser chooser = new JFileChooser();

        int result = chooser.showOpenDialog(this);
        if (result == JFileChooser.APPROVE_OPTION)
        {
            File f = chooser.getSelectedFile();
            name5170.setText(f.getAbsolutePath());
            load5170(false);
        }
    }//GEN-LAST:event_browse5170ActionPerformed

    private void browse5271ActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_browse5271ActionPerformed
        JFileChooser chooser = new JFileChooser();

        int result = chooser.showOpenDialog(this);
        if (result == JFileChooser.APPROVE_OPTION)
        {
            File f = chooser.getSelectedFile();
            name5271.setText(f.getAbsolutePath());
            load5271(false);
        }

    }//GEN-LAST:event_browse5271ActionPerformed

    private void createHybrid1KActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_createHybrid1KActionPerformed
        doCreateHybrid1K();
    }//GEN-LAST:event_createHybrid1KActionPerformed

    private void saveHybrid1KActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_saveHybrid1KActionPerformed
        if (rom1k == null)
        {
            return;
        }

        JFileChooser chooser = new JFileChooser();

        int result = chooser.showSaveDialog(this);
        if (result == JFileChooser.APPROVE_OPTION)
        {
            saveRom(chooser.getSelectedFile(), rom1k);
            name1k.setText(chooser.getSelectedFile().getAbsolutePath());
        }

    }//GEN-LAST:event_saveHybrid1KActionPerformed

    private void launch5170ActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_launch5170ActionPerformed
        launch(rom5170);
    }//GEN-LAST:event_launch5170ActionPerformed

    private void launch5271ActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_launch5271ActionPerformed
        launch(rom5271);
    }//GEN-LAST:event_launch5271ActionPerformed

    private void launch1kActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_launch1kActionPerformed
        launch(rom1k);
    }//GEN-LAST:event_launch1kActionPerformed

    private void stopMCUActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_stopMCUActionPerformed
        if (mcu == null || !mcu.isRunning())
        {
            return;
        }
        mcu.stopMCU(false);
    }//GEN-LAST:event_stopMCUActionPerformed

    private void createHybrid2KActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_createHybrid2KActionPerformed
        doCreateHybrid2K();
    }//GEN-LAST:event_createHybrid2KActionPerformed

    private void saveHybrid2KActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_saveHybrid2KActionPerformed
        if (rom2k == null)
        {
            return;
        }

        JFileChooser chooser = new JFileChooser();

        int result = chooser.showSaveDialog(this);
        if (result == JFileChooser.APPROVE_OPTION)
        {
            saveRom(chooser.getSelectedFile(), rom2k);
            name2k.setText(chooser.getSelectedFile().getAbsolutePath());
        }
    }//GEN-LAST:event_saveHybrid2KActionPerformed

    private void launch2kActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_launch2kActionPerformed
        launch(rom2k);
    }//GEN-LAST:event_launch2kActionPerformed

    private void browse1kActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_browse1kActionPerformed
        JFileChooser chooser = new JFileChooser();

        int result = chooser.showOpenDialog(this);
        if (result == JFileChooser.APPROVE_OPTION)
        {
            File f = chooser.getSelectedFile();
            name1k.setText(f.getAbsolutePath());
            load1k(false);
        }
        // TODO add your handling code here:
    }//GEN-LAST:event_browse1kActionPerformed

    private void browse2kActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_browse2kActionPerformed
        JFileChooser chooser = new JFileChooser();

        int result = chooser.showOpenDialog(this);
        if (result == JFileChooser.APPROVE_OPTION)
        {
            File f = chooser.getSelectedFile();
            name2k.setText(f.getAbsolutePath());
            load2k(false);
        }
    }//GEN-LAST:event_browse2kActionPerformed

    private void custom2kActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_custom2kActionPerformed

        if (rom2k == null || rom2k.length < 2048)
        {
            return;
        }

        byte [] romCopy = new byte[rom2k.length];
        System.arraycopy(rom2k, 0, romCopy, 0, rom2k.length);
        Customize2kPanel cp = new Customize2kPanel(romCopy);


        if (JOptionPane.showConfirmDialog(this, cp, "Customize ROM image",
                JOptionPane.OK_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE) == JOptionPane.OK_OPTION)
        {
            rom2k = romCopy;
            RomPatcher.checksum(rom2k, 0xDF);
        }


    }//GEN-LAST:event_custom2kActionPerformed

    private void custom1kActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_custom1kActionPerformed
        if (rom1k == null || rom1k.length < 1024)
        {
            return;
        }

        byte [] romCopy = new byte[rom1k.length];
        System.arraycopy(rom1k, 0, romCopy, 0, rom1k.length);
        Customize1kPanel cp = new Customize1kPanel(romCopy);


        if (JOptionPane.showConfirmDialog(this, cp, "Customize ROM image",
                JOptionPane.OK_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE) == JOptionPane.OK_OPTION)
        {
            rom1k = romCopy;
            RomPatcher.checksum(rom1k, 0xDF);
        }
    }//GEN-LAST:event_custom1kActionPerformed


    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JButton browse1k;
    private javax.swing.JButton browse2k;
    private javax.swing.JButton browse5170;
    private javax.swing.JButton browse5271;
    private javax.swing.JButton createHybrid1K;
    private javax.swing.JButton createHybrid2K;
    private javax.swing.JButton custom1k;
    private javax.swing.JButton custom2k;
    private javax.swing.JLabel label5170;
    private javax.swing.JLabel label5271;
    private javax.swing.JLabel labelHybrid1K;
    private javax.swing.JLabel labelHybrid2K;
    private javax.swing.JButton launch1k;
    private javax.swing.JButton launch2k;
    private javax.swing.JButton launch5170;
    private javax.swing.JButton launch5271;
    private javax.swing.JLabel mcuStatus;
    private javax.swing.JTextField name1k;
    private javax.swing.JTextField name2k;
    private javax.swing.JTextField name5170;
    private javax.swing.JTextField name5271;
    private javax.swing.JButton saveHybrid1K;
    private javax.swing.JButton saveHybrid2K;
    private javax.swing.JButton stopMCU;
    // End of variables declaration//GEN-END:variables

}
