/* KeyboardHost: Copyright 2011 John Elliott <jce@seasip.demon.co.uk>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *   derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package info.seasip.keyboardhost;

/** Used to create alternative ROM images for capacitative IBM keyboards.
 *
 * @author John Elliott
 */
public class RomPatcher
{

    /** Make the checksum come out right.
     *
     * @param rom The ROM image to sum.
     * @param offset The byte at this address will be altered to make the
     *              8-bit sum of all bytes in the ROM be zero.
     */
    public static void checksum(byte [] rom, int offset)
    {
        byte sum = 0;

        rom[offset] = 0;
        for (int n = 0; n < rom.length; n++)
        {
            sum += rom[n];
        }
        rom[offset] = (byte)(0x100 - (sum & 0xFF));
    }

         /** Create a 1k hybrid ROM image.
     *
     * @param rom5170 ROM image from a PC/AT keyboard controller
     * @param rom5271 ROM image from a 3270PC keyboard controller
     * @return The new ROM image
     */
    public static byte[] createHybrid1k(byte [] rom5170, byte [] rom5271)
    {
        int n;
        byte [] result = new byte[1024];

        // Start with the AT image, and
        // copy the key sensitivity table from the 3270PC.
        System.arraycopy(rom5170, 0,     result, 0,    1024);
        System.arraycopy(rom5271, 0x300, result, 0x300,  64);

// Replace the default set 3 translation (02->83, 7F->84) with a lookup table
// in the last 128 bytes of ROM. The instruction before this one is JB7, so
// we know A is between 00h and 7Fh.
//
        result[0x5D] = ORL_A;       // ORL A, 0x80      ;Base of trans table
        result[0x5E] = (byte)0x80;
        result[0x5F] = MOVP3_A_atA; // MOVP3 A, @A      ;Translate scancode
        result[0x60] = JUMP0;
        result[0x61] = (byte)0x71;
        for (n = 0x62; n < 0x71; n++)
        {
            result[n] = 0;
        }


        // Free up some space for the scancode translation. Move a
        // buffer-clearing subroutine from 03B3-03C3, to 02EB-02FB
        System.arraycopy(rom5170, 0x3B3, result, 0x2EB, 17);
        result[0x2F4] = (byte)0xF1;
        result[0xA5] = CALL2;
        result[0xA6] = (byte)0xEB;
        result[0x102] = JUMP2;
        result[0x103] = (byte)0xEB;

        // Move buffer-push function from 039B-03B2, to 01CE-01E5
        System.arraycopy(rom5170, 0x39B, result, 0x1CE, 24);
        result[0x1D1] = (byte)0xD6;
        result[0x1DC] = (byte)0xE5;
        result[0x1DF] = (byte)0xE2;
        result[0x23D] = CALL1;
        result[0x23E] = (byte)0xCE;
        result[0x377] = CALL1;
        result[0x378] = (byte)0xCE;

        // Move scancode-translate function from 038F-039A to 01E6-01F1
        System.arraycopy(rom5170, 0x38F, result, 0x1E6, 12);
        result[0x1EB] = (byte)0xF1;
        result[0x1ED] = (byte)0xEF;
        result[0x356] = CALL1;
        result[0x357] = (byte)0xE6;
        result[0x361] = CALL1;
        result[0x362] = (byte)0xE6;
        result[0x36E] = CALL1;
        result[0x36F] = (byte)0xE6;
        result[0x375] = CALL1;
        result[0x376] = (byte)0xE6;

        // Move function from 037B-0388 to 01F2-01FF
        System.arraycopy(rom5170, 0x37B, result, 0x1F2, 14);
        result[0x1FD] = (byte)0xFA;
        result[0x36C] = CALL1;
        result[0x36D] = (byte)0xF2;

        // Move code fragment from 0389-038E to 0x62
        System.arraycopy(rom5170, 0x389, result, 0x62, 6);
        result[0x355] = (byte)0x7B;
        result[0x37B] = JUMP0;
        result[0x37C] = (byte)0x62;

        // Tidy up
        result[0x37D] = 0;
        result[0x37E] = 0;
        result[0x37F] = 0;


        // Init the translation table
        for (n = 0; n < 0x80; n++)
        {
            result[0x380 + n] = (byte)n;
        }
        // Default Set 3 translations.
        result[0x382] = (byte)0x83;   // 02 -> 83
        result[0x3FF] = (byte)0x84;   // 7F -> 84

        // Reassign the first 12 PF keys.
        result[0x387] = (byte)0x05;   // 07 -> 05, F1
        result[0x38F] = (byte)0x06;   // 0F -> 06, F2
        result[0x397] = (byte)0x04;   // 17 -> 04, F3
        result[0x39F] = (byte)0x0C;   // 1F -> 0C, F4
        result[0x3A7] = (byte)0x03;   // 27 -> 03, F5
        result[0x3AF] = (byte)0x0B;   // 2F -> 0B, F6
        result[0x3B7] = (byte)0x83;   // 37 -> 83, F7
        result[0x3BF] = (byte)0x0A;   // 3F -> 0A, F8
        result[0x3C7] = (byte)0x01;   // 47 -> 01, F9
        result[0x3CF] = (byte)0x09;   // 4F -> 09, F10
        result[0x3D6] = (byte)0x78;   // 56 -> 78, F11
        result[0x3DE] = (byte)0x07;   // 5E -> 07, F12

        // Reassign the second 12 PF keys.
        result[0x388] = (byte)0x1F;   // 08 -> 1F, F13
        result[0x390] = (byte)0x27;   // 10 -> 27, F14
        result[0x398] = (byte)0x2F;   // 18 -> 2F, F15
        result[0x3A0] = (byte)0x5E;   // 20 -> 5E, F16
        result[0x3A8] = (byte)0x08;   // 28 -> 08, F17
        result[0x3B0] = (byte)0x10;   // 30 -> 10, F18
        result[0x3B8] = (byte)0x18;   // 38 -> 18, F19
        result[0x3C0] = (byte)0x20;   // 40 -> 20, F20
        result[0x3C8] = (byte)0x28;   // 28 -> 08, F21
        result[0x3D0] = (byte)0x30;   // 30 -> 10, F22
        result[0x3D7] = (byte)0x38;   // 37 -> 18, F23
        result[0x3DF] = (byte)0x40;   // 3F -> 20, F24

        // Sort out Caps Lock, Ctrl and Alt
        result[0x391] = (byte)0x14;   // 11 -> 14, Ctrl
        result[0x394] = (byte)0x58;   // 14 -> 58, Caps Lock
        result[0x399] = (byte)0x11;   // 19 -> 11, Alt
        result[0x3B9] = (byte)0x11;   // 39 -> 11, Alt
        result[0x3D8] = (byte)0x14;   // 58 -> 14, Ctrl

        // Sort out backslashes etc.
        result[0x393] = (byte)0x61;   // 13 -> 61, key beside left shift
        result[0x3D3] = (byte)0x5D;   // 53 -> 5D, key beside vertical Enter
        result[0x3DC] = (byte)0x5D;   // 5C -> 5D, key above horizontal Enter

        // These keys in the middle block return scancodes that, when
        // translated, get interpreted by the keyboard controller as protocol
        // scancodes. Reassign some of the ones displaced from the function
        // keys.

        result[0x3E5] = (byte)0x0F; // 65 -> 0F
        result[0x3ED] = (byte)0x17; // 6D -> 17


        // Reassign the left block with scancodes displaced from elsewhere.
        result[0x381] = (byte)0x5C;   // L9:  01 -> 5C
        result[0x382] = (byte)0x56;   // L7:  02 -> 56
        result[0x383] = (byte)0x39;   // L5:  03 -> 53, CrSel
        result[0x384] = (byte)0x53;   // L3:  04 -> 39, ExSel
        result[0x385] = (byte)0x84;   // L1:  05 -> 84, SysRQ
        result[0x386] = (byte)0x76;   // L2:  06 -> 76, Escape
        result[0x389] = (byte)0x48;   // L10: 09 -> 48
        result[0x38A] = (byte)0x5F;   // L8:  0A -> 5F, Clear
        result[0x38B] = (byte)0x50;   // L6:  0B -> 50
        result[0x38C] = (byte)0x19;   // L4:  0C -> 19, Attn

        // Because some keys assigned are duplicates, we are left with no keys
        // producing the following scancodes. The ones in square brackets
        // get translated by the keyboard controller to protocol scancodes
        // so aren't a problem anyway:
        // 37 3F [47] [4F] 57 [65] [6D]


        // Free space in this ROM image:
        // 0x0068-0x0070
        // 0x00E0-0x00EC
        // 0x02FC-0x02FF
        // 0x037D-0x037F

        // Finally, fix the checksum.
        checksum(result, 0xDF);
        return result;
    }

    /** Create a 1k hybrid ROM image.
     *
     * @param rom5170 ROM image from a PC/AT keyboard controller
     * @param rom5271 ROM image from a 3270PC keyboard controller
     * @return The new ROM image
     */
    public static byte[] createSimpleHybrid1k(byte [] rom5170, byte [] rom5271)
    {
        byte [] result = new byte[1024];

        // The 1k patch is very simple: Start with the AT image, and
        // copy the key sensitivity table from the 3270PC.
        System.arraycopy(rom5170, 0,     result, 0,    1024);
        System.arraycopy(rom5271, 0x300, result, 0x300,  64);

        // Then fix the checksum.
        checksum(result, 0xDF);
        return result;
    }


    /** Create a 2k hybrid ROM image.
     *
     * @param rom5170 ROM image from a PC/AT keyboard controller
     * @param rom5271 ROM image from a 3270PC keyboard controller
     * @return The new ROM image
     */
    public static byte[] createHybrid2k(byte [] rom5170, byte [] rom5271)
    {
        int n;
        byte [] result = new byte[2048];

        // The patches below are in the approximate order that I wrote them,
        // rather than any sort of logical sequence. They have been
        // hand-assembled, which means that here and there you'll find
        // non-functional bits left in to avoid changing all the offsets.

        // Start the same way as the 1k ROM: Copy the 5170 firmware, and
        // superimpose the 5271 key sensitivity table.
        //
        System.arraycopy(rom5170, 0,     result, 0,    1024);
        System.arraycopy(rom5271, 0x300, result, 0x300,  64);

        // Now we start expanding. Firstly, extend the ROM checksum to cover all
        // 2k of controller ROM. This requires code in the four extra pages.
        result[0x7FD] = MOVP_A_atA; // MOVP A, @A
        result[0x7FE] = JUMP2;      // JMP ADD_CKSUM
        result[0x7FF] = ADD_CKSUM;
        result[0x6FD] = MOVP_A_atA; // MOVP A, @A
        result[0x6FE] = JUMP2;      // JMP ADD_CKSUM
        result[0x6FF] = ADD_CKSUM;
        result[0x5FD] = MOVP_A_atA; // MOVP A, @A
        result[0x5FE] = JUMP2;      // JMP ADD_CKSUM
        result[0x5FF] = ADD_CKSUM;
        result[0x4FD] = MOVP_A_atA; // MOVP A, @A
        result[0x4FE] = JUMP2;      // JMP ADD_CKSUM
        result[0x4FF] = ADD_CKSUM;
        // All this new code is invoked at CKSUM2K
        result[0x119] = JUMP7;
        result[0x11A] = CKSUM2K;    // JMP 7F1 (replaces CALL 0D9)

        // Add on checksum bytes for pages 0, 4, 5, 6, 7. Replaces a call to
        // CKSUM_P0, that just added on the byte for page 0.
// CKSUM2K:
        result[0x7F1] = CALL0;
        result[0x7F2] = CKSUM_P0;    // CALL 0D9
        result[0x7F3] = CALL4;
        result[0x7F4] = CKSUM_P7;    // CALL 4FD
        result[0x7F5] = CALL5;
        result[0x7F6] = CKSUM_P7;    // CALL 5FD
        result[0x7F7] = CALL6;
        result[0x7F8] = CKSUM_P7;    // CALL 6FD
        result[0x7F9] = CALL7;
        result[0x7FA] = CKSUM_P7;    // CALL 7FD
        result[0x7FB] = JUMP1;
        result[0x7FC] = (byte)0x1B;  // JMP 011B

        // Increase the RAM test from 64 bytes to 128.
        result[0x10B] = 0x7F;       // Last address to test.

// OK, we have now adjusted for the new size of the ROM and the RAM: this is
// now an 8049 firmware, not an 8048 firmware.
//
// Next, move the key decoder from page 3 to page 7. The MOVP3 instruction
// means that page 3 is best left to data tables.

        System.arraycopy(result, 0x340, result, 0x740, 0xB0);
// Blank out what was in page 3.
        for (n = 0x340; n < 0x3F0; n++)
        {
            result[n] = 0;
        }
        // Now replace all references to code in page 3 with page 7.
        result[0x0A5] = CALL7; // CALL 3xx -> CALL 7xx
        result[0x102] = JUMP7; // JMP  3xx -> JMP  7xx
        result[0x23D] = CALL7; // CALL 3xx -> CALL 7xx
        result[0x2A3] = JUMP7; // JMP  3xx -> JMP  7xx
        result[0x756] = CALL7; // CALL 3xx -> CALL 7xx
        result[0x761] = CALL7; // CALL 3xx -> CALL 7xx
        result[0x765] = JUMP7; // JMP  3xx -> JMP  7xx
        result[0x76C] = CALL7; // CALL 3xx -> CALL 7xx
        result[0x76E] = CALL7; // CALL 3xx -> CALL 7xx
        result[0x775] = CALL7; // CALL 3xx -> CALL 7xx
        result[0x777] = CALL7; // CALL 3xx -> CALL 7xx
        result[0x779] = JUMP7; // JMP  3xx -> JMP  7xx

///////////////////////////////////////////////////////////////////////////////
//
// Now add support for all three scancode sets. Replace the inline translation
// code at 0x5B with three different routines for set 1 / 2 / 3.
//
        result[0x5B]  = JUMP5; // JMP 580
        result[0x5C]  = XLT_SWITCH;

        // Do the set 3 translation at 700h. This is also used
        // as a final translation stage by set 2.
        //
        // Code 0x02 is mapped to 0x83
        // Code 0x7F is mapped to 0x84
        //
        // I don't know why this is done. Maybe IDENT and SPACE had some
        // special meaning on 3270 terminals?
        //
        // Enter with A = scancode.
        //
        result[0x700]  = XRL_A;      // XRL A, 2    2 -> 0x83
        result[0x701]  = (byte)0x02;
        result[0x702]  = JNZ;        // JNZ 708h
        result[0x703]  = SET3_NOT02;
        result[0x704]  = MOV_A;      // MOV A, 83h
        result[0x705]  = (byte)0x83;
        result[0x706]  = JUMP0;      // JMP DO_SEND_SC
        result[0x707]  = DO_SEND_SC;
//
// SET3_NOT02:
        result[0x708]  = XRL_A;      // XRL A, 7Dh  ; 2 ^ 7D = 7F
        result[0x709]  = (byte)0x7D;
        result[0x70A]  = JNZ;        // JNZ 710h
        result[0x70B]  = SET3_NOT7F;
        result[0x70C]  = MOV_A;      // MOV A, 84h
        result[0x70D]  = (byte)0x84;
        result[0x70E]  = JUMP0;      // JMP DO_SEND_SC
        result[0x70F]  = DO_SEND_SC;
//
// SET3_NOT7F:
        result[0x710]  = XRL_A;      // XRL A, 7FH  ; Get original value back
        result[0x711]  = (byte)0x7F;
        result[0x712]  = JUMP0;      // JMP DO_SEND_SC
        result[0x713]  = DO_SEND_SC;
//
// The set 2 translation is by far the most involved, and occupies most of
// page 6.
//
// Enter with the scancode to translate at CUR_SC. The translation table is
// stored from 0340h to 03BFh, indexed by set 3 scancode. Each byte means:
//      00-7F:    Send that scancode (with 02 -> 83, 7F -> 84).
//      80:       Ignore key
//      81-DF:    Send E0 followed by (value & 0x7F), with usual 02->83, 7F->84
//      E0:       Send Break, E0 7E E0 F0 7E
//      E1:       Send Pause, E1 14 77 E1 F0 14 F0 77
//      E2-FF:    As 81-DF.
//
// XLT_SET2:
        result[0x600] = MOV_R0;      // MOV R0, CUR_SC
        result[0x601] = CUR_SC;
        result[0x602] = MOV_A_atR0;  // MOV A, @R0   ; @R0 is set 3 scancode
        result[0x603] = ANL_A;       // ANL A, 7Fh   ; Mask off 'key-up' flag
        result[0x604] = (byte)0x7F;
        result[0x605] = ADD_A;       // ADD A, TABLE_SET2
        result[0x606] = TABLE_SET2;
        result[0x607] = MOVP3_A_atA; // MOVP3 A, @A   ;A is set 2 scancode
//
// An inelegant fix. I had originally intended 00 to be ignored, like 80.
// But since the keyboard can actually send 00 (buffer full) it needs to be
// treated as a normal scancode.
//
        result[0x608] = JUMP6; //JNZ;// JNZ SET2_NOT0 ;If set 2 scancode is 0,
				     //               ;return at once
        result[0x609] = SET2_NOT0;
// SET2_SKIP:
        result[0x60A] = JUMP0;       // JMP SWALLOW_SC ; Ignore scancode
        result[0x60B] = SWALLOW_SC;

// SET2_NOT0:
        result[0x60C] = JB7;         // JB7 SET2_EXT   ;Bit 7 set for specials
        result[0x60D] = SET2_EXT;
//
// Bit 7 is not set; we just have to translate this code, and send it.
//
        result[0x60E] = XCH_A_atR0; // XCH A, @R0      ;A = original scancode
        result[0x60F] = ANL_A;      // ANL A, 80h      ;A = key-up flag
        result[0x610] = (byte)0x80;
        result[0x611] = ORL_A_atR0; // ORL A, @R0      ;A = new scancode | up
        result[0x612] = JB7;        // JB7 SET2_UP
        result[0x613] = SET2_UP;
        result[0x614] = JUMP7;      // JMP 700h        ;Chain to set 3 for
        result[0x615] = XLT_SET3;   //                 ;02 -> 83 and 7F -> 84
//
// This is a key-up event. Send the key-up code (0xF0) and reset the key-up
// flag on the scancode in the buffer.
//
// SET2_UP:
        result[0x616] = JUMP0;      // JMP 07Ah        ;Key-up. Send 0xF0
        result[0x617] = DO_SEND_F0; //                 ;and leave scancode

// OK, this is a special-case key. Check for the one-offs: E1, E0 and 80.
//
// SET2_EXT:
        result[0x618] = XRL_A;      // XRL A,0xE1    ;Was it E1?
        result[0x619] = (byte)0xE1;
        result[0x61A] = JZ;         // JZ SET2_PAUSE
        result[0x61B] = SET2_PAUSE;

        result[0x61C] = XRL_A;      // XRL A,0x01    ;Was it E0?
        result[0x61D] = (byte)0x01;
        result[0x61E] = JZ;         // JZ SET2_BREAK
        result[0x61F] = SET2_BREAK;

        result[0x620] = XRL_A;      // XRL A,0x60   ;Was it 80?
        result[0x621] = (byte)0x60;
        result[0x622] = JZ;         // SET2_SKIP    ;If so, skip it
        result[0x623] = SET2_SKIP;
        result[0x624] = XRL_A;      // XRL A,0x80   ;Undo all the XORing: A is
        result[0x625] = (byte)0x80; //              ;what it was before.

// OK. It's of the form E0 followed by a normal scancode. A is
// (scancode | 0x80). R0 points to key-up flag.
// Byte at FLAG_E0 has bit 7 set if E0 sent, clear if E0 not sent
//
        result[0x626] = INC_R0;     // INC R0       ;R0 -> FLAG_E0
        result[0x627] = XRL_A_atR0; // XRL A, @R0   ;If bit 7 of A is now set,
                                    //              ;E0 has not been sent
        result[0x628] = JB7;        // JB7 DO_SEND_E0
        result[0x629] = DO_SEND_E0;
        result[0x62A] = DEC_R0;     // DEC R0
        result[0x62B] = XCH_A_atR0; // XCH A, @R0   ;A = original code
        result[0x62C] = ANL_A;      // ANL A, 80h   ;A = key-up flag
        result[0x62D] = (byte)0x80;
        result[0x62E] = ORL_A_atR0; // ORL A, @R0   ;A = (scancode | key-up)
        result[0x62F] = JB7;        // JB7 SET2_SEND_F0 ; If bit 7 set, send
        result[0x630] = SET2_SEND_F0; //                ; 0xF0 and continue
//
// Bit 7 not set, so A is just the scancode. Send it and reset the E0
// indicator.
//
        result[0x631] = INC_R0;   // INC R0         ;R0 -> FLAG_E0
        result[0x632] = MOV_atR0; // MOV @R0, 0
        result[0x633] = (byte)0x00;
        result[0x634] = JUMP7;    // JMP 700h   ;Now hand off to the set 3
        result[0x635] = XLT_SET3; //            ;translator for 02->83, 7F->84
//
// The E0 has not been sent, so send it. Leave the scancode in the buffer so
// it will be picked up again for the rest of the sequence.
//
// DO_SEND_E0:
        result[0x636] = MOV_A;    // MOV A, 0E0h
        result[0x637] = (byte)0xE0;
        result[0x638] = CALL1;    // CALL 0183h
        result[0x639] = TRANSMIT;
        result[0x63A] = JF1;      // JF1 642h
        result[0x63B] = SET2_SENDERR;
//
// We have sent the E0, so set the "E0 has been sent" flag.
//
        result[0x63C] = MOV_R0; // MOV R0, FLAG_E0
        result[0x63D] = FLAG_E0;
        result[0x63E] = MOV_atR0; // MOV @R1, 80h ;E0 has been sent
        result[0x63F] = (byte)0x80;
// SET2_SCAN:
        result[0x640] = JUMP2;    // JMP 205h
        result[0x641] = SCAN_KB;
//
// E0 was not sent successfully
//
// SET2_SENDERR:
        result[0x642] = JUMP0; // JMP 077h
        result[0x643] = (byte)0x77;
//
// This is a key-up: Send F0 and reset 'key-up' flag.
//
        result[0x644] = JUMP0; // JMP 07Eh
        result[0x645] = (byte)0x7e;
//
// Handle 0xE1: Pause key.
//
// Byte sequence to send is: E1 14 77 E1 F0 14 F0 77 on down, nothing on up.
// R0 -> key-up flag.
//
// SET2_PAUSE:
        result[0x646] = MOV_A_atR0; // MOV A, @R0
        result[0x647] = JB7;        // JB7 SET2_SKIP  ;If key-up, do nothing.
        result[0x648] = SET2_SKIP;
//
// This is part of an unsuccessful idea to make keys do different things
// depending what shifts are pressed. The idea was that Ctrl+Pause would
// be interpreted as Break. That didn't work out so Break has to be put on
// a separate key.
//
        result[0x649] = MOV_R0;       // MOV R0, SHIFT_STATE
        result[0x64A] = (byte)SHIFT_STATE;
        result[0x64B] = MOV_A_atR0;   // MOV A, @R0
        result[0x64C] = JB2;          // JB2 0x668    ; Left Ctrl down
        result[0x64D] = SET2_BREAK;
        result[0x64E] = JB3;          // JB3 0x668    ; Right Ctrl down
        result[0x64F] = SET2_BREAK;
//
// Pause with Ctrl not pressed...
//
        result[0x650] = MOV_R0;       // MOV R0, FLAG_E0
        result[0x651] = FLAG_E0;
        result[0x652] = MOV_A_atR0;   // MOV A, @R0         ;If we're not
        result[0x653] = JNZ;          // JNZ SET2_PAUSE1    ;midsequence, point
        result[0x654] = SET2_PAUSE1;  //                    ;R0 at the start of
        result[0x655] = MOV_atR0;     // MOV @R0, PAUSE_SC2 ;the byte stream
        result[0x656] = PAUSE_SC2;
// SET2_PAUSE1:
        result[0x657] = MOV_A_atR0;   // MOV A, @R0         ;A->next byte
        result[0x658] = MOVP_A_atA;   // MOVP A, @A         ;Get it
        result[0x659] = CALL1;        // CALL TRANSMIT      ;Transmit it
        result[0x65A] = TRANSMIT;
        result[0x65B] = JF1;          // JF1 SET2_SENDERR   ;Wasn't sent
        result[0x65C] = SET2_SENDERR;
        result[0x65D] = MOV_R0;       // MOV R0, FLAG_E0
        result[0x65E] = FLAG_E0;
        result[0x65F] = INC_atR0;     // INC @R0       ;R0->Next byte in
        result[0x660] = MOV_A_atR0;   // MOV A, @R0    ;sequence
        result[0x661] = MOVP_A_atA;   // MOVP A, @A    ;Get next byte.
        result[0x662] = JNZ;          // JNZ SET2_SCAN ;Not the last; continue.
        result[0x663] = SET2_SCAN;
        result[0x664] = MOV_atR0;     // MOV @R0, 00h  ;Last byte done; reset
        result[0x665] = (byte)0x00;   //               ;the flag
        result[0x666] = JUMP0;        // JMP SWALLOW_SC ;& Remove scancode
        result[0x667] = SWALLOW_SC;   //                ; from the buffer

// Handle 0xE0, break
SET2_BREAK:
        result[0x668] = MOV_A_atR0;       // MOV A, @R0    ;Is it key-up?
        result[0x669] = JB7;              // JB7 SET2_SKIP ;If so, do nothing
        result[0x66A] = SET2_SKIP;
        result[0x66B] = MOV_R0;           // MOV R0, FLAG_E0
        result[0x66C] = FLAG_E0;
        result[0x66D] = MOV_A_atR0;       // MOV A, @R0
        result[0x66E] = JNZ;              // JNZ SET2_PAUSE1 ;First byte in
        result[0x66F] = SET2_PAUSE1;      //                 ;sequence?
        result[0x670] = MOV_atR0;         // MOV @R0, BREAK_SC2
        result[0x671] = BREAK_SC2;
        result[0x672] = JUMP6;            // JMP SET2_PAUSE1
        result[0x673] = SET2_PAUSE1;
//
// End of Set 2 translation code (the key sequences generated by
// BREAK and PAUSE come later).
//
// Next, Set 1. This works by translating to set 2, and then going through the
// standard Set 2-> Set 1 table:
// <http://www.win.tue.nl/~aeb/linux/kbd/scancodes-10.html#ss10.3>
//
	result[0x500] = (byte)0xff;	// 00 -> FF
	result[0x501] = (byte)0x43;	// 01 -> 43
	result[0x502] = (byte)0x41;	// 02 -> 41 (83 -> 41)
	result[0x503] = (byte)0x3F;	// 03 -> 3F
	result[0x504] = (byte)0x3D;	// 04 -> 3D
	result[0x505] = (byte)0x3B;	// 05 -> 3B
	result[0x506] = (byte)0x3C;	// 06 -> 3C
	result[0x507] = (byte)0x58;	// 07 -> 58
	result[0x508] = (byte)0x64;	// 08 -> 64
	result[0x509] = (byte)0x44;	// 09 -> 44
	result[0x50a] = (byte)0x42;	// 0A -> 42
	result[0x50b] = (byte)0x40;	// 0B -> 40
	result[0x50c] = (byte)0x3E;	// 0C -> 3E
	result[0x50d] = (byte)0x0F;	// 0D -> 0F
	result[0x50e] = (byte)0x29;	// 0E -> 29
	result[0x50f] = (byte)0x59;	// 0F -> 59
	result[0x510] = (byte)0x65;	// 10 -> 65
	result[0x511] = (byte)0x38;	// 11 -> 38
	result[0x512] = (byte)0x2A;	// 12 -> 2A
	result[0x513] = (byte)0x70;	// 13 -> 70
	result[0x514] = (byte)0x1D;	// 14 -> 1D
	result[0x515] = (byte)0x10;	// 15 -> 10
	result[0x516] = (byte)0x02;	// 16 -> 02
	result[0x517] = (byte)0x5A;	// 17 -> 5A
	result[0x518] = (byte)0x66;	// 18 -> 66
	result[0x519] = (byte)0x71;	// 19 -> 71
	result[0x51a] = (byte)0x2c;	// 1A -> 2C
	result[0x51b] = (byte)0x1f;	// 1B -> 1F
	result[0x51c] = (byte)0x1e;	// 1C -> 1E
	result[0x51d] = (byte)0x11;	// 1D -> 11
	result[0x51e] = (byte)0x03;	// 1E -> 03
	result[0x51f] = (byte)0x5B;	// 1F -> 5B
	result[0x520] = (byte)0x67;	// 20 -> 67
	result[0x521] = (byte)0x2e;	// 21 -> 2E
	result[0x522] = (byte)0x2d;	// 22 -> 2D
	result[0x523] = (byte)0x20;	// 23 -> 20
	result[0x524] = (byte)0x12;	// 24 -> 12
	result[0x525] = (byte)0x05;	// 25 -> 05
	result[0x526] = (byte)0x04;	// 26 -> 04
	result[0x527] = (byte)0x5C;	// 27 -> 5C
	result[0x528] = (byte)0x68;	// 28 -> 68
	result[0x529] = (byte)0x39;	// 29 -> 39
	result[0x52a] = (byte)0x2F;	// 2A -> 2F
	result[0x52b] = (byte)0x21;	// 2B -> 21
	result[0x52c] = (byte)0x14;	// 2C -> 14
	result[0x52d] = (byte)0x13;	// 2D -> 13
	result[0x52e] = (byte)0x06;	// 2E -> 06
	result[0x52f] = (byte)0x5D;	// 2F -> 5D
	result[0x530] = (byte)0x69;	// 30 -> 69
	result[0x531] = (byte)0x31;	// 31 -> 31
	result[0x532] = (byte)0x30;	// 32 -> 30
	result[0x523] = (byte)0x23;	// 33 -> 23
	result[0x534] = (byte)0x22;	// 34 -> 22
	result[0x535] = (byte)0x15;	// 35 -> 15
	result[0x536] = (byte)0x07;	// 36 -> 07
	result[0x537] = (byte)0x5E;	// 37 -> 5E
	result[0x538] = (byte)0x6A;	// 38 -> 6A
	result[0x539] = (byte)0x72;	// 39 -> 72
	result[0x53a] = (byte)0x32;	// 3A -> 32
	result[0x53b] = (byte)0x24;	// 3B -> 24
	result[0x53c] = (byte)0x16;	// 3C -> 16
	result[0x53d] = (byte)0x08;	// 3D -> 08
	result[0x53e] = (byte)0x09;	// 3E -> 09
	result[0x53f] = (byte)0x5F;	// 3F -> 5F
	result[0x540] = (byte)0x6B;	// 40 -> 6B
	result[0x541] = (byte)0x33;	// 41 -> 33
	result[0x542] = (byte)0x25;	// 42 -> 25
	result[0x543] = (byte)0x17;	// 43 -> 17
	result[0x544] = (byte)0x18;	// 44 -> 18
	result[0x545] = (byte)0x0B;	// 45 -> 0B
	result[0x546] = (byte)0x0A;	// 46 -> 0A
	result[0x547] = (byte)0x60;	// 47 -> 60
	result[0x548] = (byte)0x6C;	// 48 -> 6C
	result[0x549] = (byte)0x34;	// 49 -> 34
	result[0x54a] = (byte)0x35;	// 4A -> 35
	result[0x54b] = (byte)0x26;	// 4B -> 26
	result[0x54c] = (byte)0x27;	// 4C -> 27
	result[0x54d] = (byte)0x19;	// 4D -> 19
	result[0x54e] = (byte)0x0C;	// 4E -> 0C
	result[0x54f] = (byte)0x61;	// 4F -> 61
	result[0x550] = (byte)0x6D;	// 50 -> 6D
	result[0x551] = (byte)0x73;	// 51 -> 73
	result[0x552] = (byte)0x28;	// 52 -> 28
	result[0x553] = (byte)0x74;	// 53 -> 74
	result[0x554] = (byte)0x1A;	// 54 -> 1A
	result[0x545] = (byte)0x0D;	// 55 -> 0D
	result[0x556] = (byte)0x62;	// 56 -> 62
	result[0x557] = (byte)0x6E;	// 57 -> 6E
	result[0x558] = (byte)0x3A;	// 58 -> 3A
	result[0x559] = (byte)0x36;	// 59 -> 36
	result[0x55a] = (byte)0x1C;	// 5A -> 1C
	result[0x55b] = (byte)0x1B;	// 5B -> 1B
	result[0x55c] = (byte)0x75;	// 5C -> 75
	result[0x55d] = (byte)0x2B;	// 5D -> 2B
	result[0x55e] = (byte)0x63;	// 5E -> 63
	result[0x55f] = (byte)0x76;	// 5F -> 76
	result[0x560] = (byte)0x55;	// 60 -> 55
	result[0x561] = (byte)0x56;	// 61 -> 56
	result[0x562] = (byte)0x77;	// 62 -> 77
	result[0x563] = (byte)0x78;	// 63 -> 78
	result[0x564] = (byte)0x79;	// 64 -> 79
	result[0x565] = (byte)0x7A;	// 65 -> 7A
	result[0x566] = (byte)0x0E;	// 66 -> 0E
	result[0x567] = (byte)0x7B;	// 67 -> 7B
	result[0x568] = (byte)0x7C;	// 68 -> 7C
	result[0x569] = (byte)0x4F;	// 69 -> 4F
	result[0x56a] = (byte)0x7D;	// 6A -> 7D
	result[0x56b] = (byte)0x4B;	// 6B -> 4B
	result[0x56c] = (byte)0x47;	// 6C -> 47
	result[0x56d] = (byte)0x7E;	// 6D -> 7E
	result[0x56e] = (byte)0x7F;	// 6E -> 7F
	result[0x56f] = (byte)0x6F;	// 6F -> 6F
	result[0x570] = (byte)0x52;	// 70 -> 52
	result[0x571] = (byte)0x53;	// 71 -> 53
	result[0x572] = (byte)0x50;	// 72 -> 50
	result[0x573] = (byte)0x4C;	// 73 -> 4C
	result[0x574] = (byte)0x4D;	// 74 -> 4D
	result[0x575] = (byte)0x48;	// 75 -> 48
	result[0x576] = (byte)0x01;	// 76 -> 01
	result[0x577] = (byte)0x45;	// 77 -> 45
	result[0x578] = (byte)0x57;	// 78 -> 57
	result[0x579] = (byte)0x4E;	// 79 -> 4E
	result[0x57A] = (byte)0x51;	// 7A -> 51
	result[0x57B] = (byte)0x4A;	// 7B -> 4A
	result[0x57C] = (byte)0x37;	// 7C -> 37
	result[0x57D] = (byte)0x49;	// 7D -> 49
	result[0x57E] = (byte)0x46;	// 7E -> 46
	result[0x57F] = (byte)0x54;	// 7F -> 54 (84 -> 54)
//
// The Set 1 translation code. It's very similar to the Set 2 code, the main
// difference being that there's no separate F0 code for key-up.
//
// XLT_SET1:
        result[0x580] = MOV_R0;         // MOV R0, CUR_SC
        result[0x581] = CUR_SC;
        result[0x582] = MOV_A_atR0;     // MOV A, @R0 ; @R0 is set 3 scancode
        result[0x583] = ANL_A;          // ANL A, 7Fh ; Mask off 'key-up' flag
        result[0x584] = (byte)0x7F;
        result[0x585] = ADD_A;          // ADD A, TABLE_SET2
        result[0x586] = TABLE_SET2;
        result[0x587] = MOVP3_A_atA;    // MOVP3 A, @A   ; A is set 2 scancode
//
// The same inelegant fix as for set 2.
//
        result[0x588] = JUMP5;  // JNZ; // JNZ SET1_NOT0
        result[0x589] = SET1_NOT0;
// SET1_SKIP:
        result[0x58A] = JUMP0;          // JMP SWALLOW_SC ; Scancode 0 ignored
        result[0x58B] = SWALLOW_SC;

// SET1_NOT0:
        result[0x58C] = JB7;            // JB7 SET1_EXT   ;Bit 7 set for
        result[0x58D] = SET1_EXT;       //                ;E0 and special keys
//
// No E0 prefix. We have a simple 1:1 scancode translation; and, unlike in
// set 2, we don't have to send the key-up prefix separately.
//
        result[0x58E] = XCH_A_atR0; // XCH A, @R0     ;@R0 is set 2 scancode
        result[0x58F] = ANL_A;      // ANL A, 80h     ;A is key-up flag
        result[0x590] = (byte)0x80;
        result[0x591] = XCH_A_atR0; // XCH A, @R0     ;@R0 is key-up flag
	result[0x592] = MOVP_A_atA; // MOVP A, @A     ;A is set 1 scancode
        result[0x593] = ORL_A_atR0; // ORL A, @R0     ;including key-up flag
        result[0x594] = JUMP0;      // JMP DO_SEND_SC ;Send it.
        result[0x595] = DO_SEND_SC;

// Check for the one-offs: 80 E0 and E1.
//
// SET1_EXT:
        result[0x596] = XRL_A;      // XRL A,0xE1
        result[0x597] = (byte)0xE1;
        result[0x598] = JZ;         // JZ SET1_PAUSE
        result[0x599] = SET1_PAUSE;

        result[0x59A] = XRL_A;      // XRL A,0x01   ;Was it E0?
        result[0x59B] = (byte)0x01;
        result[0x59C] = JZ;         // JZ SET1_BREAK
        result[0x59D] = SET1_BREAK;

        result[0x59E] = XRL_A;      // XRL A,0x60   ;Was it 80?
        result[0x59F] = (byte)0x60;
        result[0x5A0] = JZ;         // JZ SET1_SKIP ;Currently just skipped
        result[0x5A1] = SET1_SKIP;
        result[0x5A2] = XRL_A;      // XRL A,0x80   ;Undo all the XORing:
        result[0x5A3] = (byte)0x80; //              ;A is what it was before.

// OK. It's E0 followed by a normal scancode. A is (scancode | 0x80). R0
// points to key-up flag.
// Byte at 7Dh has bit 7 set if E0 sent, clear if E0 not sent

        result[0x5A4] = INC_R0;     // INC R0       ;R0 -> FLAG_E0
        result[0x5A5] = XRL_A_atR0; // XRL A, @R0   ;If bit 7 of A is now set,
                                    //              ;E0 has not been sent
        result[0x5A6] = JB7;        // JB7 SET1_SENDE0
        result[0x5A7] = SET1_SENDE0;//		    ;A is set 2 scancode
        result[0x5A8] = DEC_R0;     // DEC R0
	result[0x5A9] = MOVP_A_atA; // MOVP A, @A   ;A is set 1 scancode
        result[0x5AA] = XCH_A_atR0; // XCH A, @R0   ;A is original scancode
        result[0x5AB] = ANL_A;      // ANL A, 80h   ;A is key-up flag
        result[0x5AC] = (byte)0x80;
        result[0x5AD] = ORL_A_atR0; // ORL A, @R0   ;A = scancode | upflag
        result[0x5AE] = INC_R0;     // INC R0
        result[0x5AF] = MOV_atR0;   // MOV @R0, 0   ;Reset 0xE0 flag
        result[0x5B0] = (byte)0x00;
        result[0x5B1] = JUMP0; 	    // JMP DO_SEND_SC ;Code translated OK.
        result[0x5B2] = DO_SEND_SC; //                ;Send it.
//
// The E0 has not been sent, so send it.
//
// SET1_SENDE0:
	result[0x5B3] = JUMP6;
	result[0x5B4] = DO_SEND_E0; // JMP DO_SEND_E0
//
// The special-case code for Pause and Break sets up the pointers and jumps
// to the send-multiple-bytes code in page 6. The code could be made a lot
// more compact if I dared to use registers other than R0; but the code I'm
// patching into only uses A and R0.
//
// SET1_PAUSE:
        result[0x5B5] = MOV_A_atR0; // MOV A, @R0
        result[0x5B6] = JB7;        // JB7 SET1_SKIP  ;If key-up, do nothing.
        result[0x5B7] = SET1_SKIP;
        result[0x5B8] = MOV_R0;     // MOV R0, FLAG_E0
        result[0x5B9] = FLAG_E0;
        result[0x5BA] = MOV_A_atR0; // MOV A, @R0
        result[0x5BB] = JNZ;        // JNZ SET1_PAUSE1 ;First byte in
        result[0x5BC] = SET1_PAUSE1;//                 ;sequence?
        result[0x5BD] = MOV_atR0;   // MOV @R0, PAUSE_SC1
        result[0x5BE] = PAUSE_SC1;
// SET1_PAUSE1:
        result[0x5BF] = JUMP6;      // JMP SET2_PAUSE1
        result[0x5C0] = SET2_PAUSE1;

// SET1_BREAK:
        result[0x5C1] = MOV_A_atR0; // MOV A, @R0
        result[0x5C2] = JB7;        // JB7 SET1_SKIP  ;If key-up, do nothing.
        result[0x5C3] = SET1_SKIP;
        result[0x5C4] = MOV_R0;     // MOV R0, FLAG_E0
        result[0x5C5] = FLAG_E0;
        result[0x5C6] = MOV_A_atR0; // MOV A, @R0
        result[0x5C7] = JNZ;        // JNZ SET1_PAUSE1 ;First byte in
        result[0x5C8] = SET1_PAUSE1;//                 ;sequence?
        result[0x5C9] = MOV_atR0;   // MOV @R0, PAUSE_SC1
        result[0x5CA] = BREAK_SC1;
        result[0x5CB] = JUMP6;      // JMP SET2_PAUSE1
        result[0x5CC] = SET2_PAUSE1;
//
// And now the code that decides which translation to use, depending on
// scancode set selected. Enter with scancode to translate in A.
//
// XLT_SWITCH:
	result[0x5CD] = MOV_R0;	    // MOV R0, CUR_SC   ;Save the current code.
	result[0x5CE] = CUR_SC;	    //
	result[0x5CF] = MOV_atR0_A; // MOV @R0, A
	result[0x5D0] = MOV_R0;	    // MOV R0, SCANSET  ;Get the set in use.
	result[0x5D1] = SCANSET;    //
	result[0x5D2] = MOV_A_atR0; // MOV A, @R0
	result[0x5D3] = DEC_A;	    // DEC A            ;Is it 1?
	result[0x5D4] = JZ;	    // JZ XLT_SET1      ;If so, jump to set 1.
	result[0x5D5] = XLT_SET1;
	result[0x5D6] = DEC_A;	    // DEC A            ;Is it 2?
	result[0x5D7] = JZ;	    // JZ J_SET2        ;If so, jump to set 2.
	result[0x5D8] = J_SET2;
	result[0x5D9] = DEC_A;	    // DEC A            ;Is it 3?
	result[0x5DA] = JZ;	    // JZ J_SET3        ;If so, jump to set 3.
	result[0x5DB] = J_SET3;
// J_SET2:
	result[0x5DC] = JUMP6;	    // JMP XLT_SET2    ;Default case: Set 2.
	result[0x5DD] = XLT_SET2;   //
// J_SET3:
	result[0x5DE] = MOV_R0;	    // MOV R0, CUR_SC  ;Set 3: Retrieve the
	result[0x5DF] = CUR_SC;	    //                 ;scancode.
	result[0x5E0] = MOV_A_atR0; // MOV A, @R0
        result[0x5E1] = JB7;        // JB7 SET3_UP     ;Is it a key-up code?
        result[0x5E2] = SET3_SENDF0; //                ;If so, send 0xF0
	result[0x5E3] = JUMP7;	    // JMP XLT_SET3    ;Otherwise, do 02->83
        result[0x5E4] = XLT_SET3;   //                 ;and 7F->84 translation.
// SET3_SENDF0:
        result[0x5E5] = JUMP0;      // JMP DO_SEND_F0  ;Inter-page jump
        result[0x5E6] = DO_SEND_F0;
//
// Set up the table used for Set 3 -> Set 2 scancode translation.
//
// Start with an identity mapping.
//
        for (n = 0; n <= 0x7F; n++)
        {
            result[0x340 + n] = (byte)n;
        }
        // Left function keys:
        result[0x345] = (byte)0x7F;   // L1  05 -> 7F   (SysRQ)
        result[0x346] = (byte)0x76;   // L2  06 -> 76   (Esc)
        result[0x344] = (byte)0xCA;   // L3  04 -> E04A (numpad slash)
        result[0x34C] = (byte)0xE1;   // L4  0C -> *E1* (Pause)
        result[0x343] = (byte)0x17;   // L5  03 -> 17   (PA1)
        result[0x34B] = (byte)0xE0;   // L6  0B -> *E0* (Break)
        result[0x342] = (byte)0xFC;   // L7  02 -> E07C (PrtSc)
        result[0x34A] = (byte)0xAF;   // L8  0A -> E02F (Menu)
        result[0x341] = (byte)0x9F;   // L9  01 -> E01F (LWin)
        result[0x349] = (byte)0xA7;   // L10 09 -> E027 (RWin)

        // 3x2 block above cursor keys
        result[0x3A2] = (byte)0xEC;   // Home (between cursors)
        result[0x3A7] = (byte)0xF0;   // Ins
        result[0x3A4] = (byte)0xF1;   // Del
        result[0x3AE] = (byte)0xEC;   // Home
        result[0x3A5] = (byte)0xE9;   // End
        result[0x3AF] = (byte)0xFD;   // Page Up
        result[0x3AD] = (byte)0xFA;   // Page Down

        // Keypad keys
        result[0x3B8] = (byte)0x63;   // Under keypad enter
        result[0x3B9] = (byte)0xDA;   // Keypad enter
        result[0x3BC] = (byte)0x79;   // Keypad +

        // Shifts
        result[0x379] = (byte)0x91;   // Right Alt
        result[0x398] = (byte)0x94;   // Right Ctrl
        result[0x351] = (byte)0x14;   // Left Ctrl
        result[0x359] = (byte)0x11;   // Left Alt
        result[0x354] = (byte)0x58;   // Caps Lock

        result[0x353] = (byte)0x61;   // Key next to left shift
        result[0x393] = (byte)0x5D;   // First backslash
        result[0x39c] = (byte)0x5D;   // Second backslash

        // Cursor keys
        result[0x3A0] = (byte)0xF2;   // Cursor down -> E072
        result[0x3A1] = (byte)0xEB;   // Cursor left -> E06B
        result[0x3A3] = (byte)0xF5;   // Cursor up -> E075
        result[0x3AA] = (byte)0xF4;   // Cursor right -> E074

        // Top function keys
        result[0x347] = (byte)0x05;   // PF1  -> F1
        result[0x348] = (byte)0x1F;   // PF13 -> F13
        result[0x34F] = (byte)0x06;   // PF2  -> F2
        result[0x350] = (byte)0x27;   // PF14 -> F14
        result[0x357] = (byte)0x04;   // PF3  -> F3
        result[0x358] = (byte)0x2F;   // PF15 -> F15
        result[0x35F] = (byte)0x0C;   // PF4  -> F4
        result[0x360] = (byte)0x5E;   // PF16 -> F16
        result[0x367] = (byte)0x03;   // PF5  -> F5
        result[0x368] = (byte)0x08;   // PF17 -> F17
        result[0x36F] = (byte)0x0B;   // PF6  -> F6
        result[0x370] = (byte)0x10;   // PF18 -> F18
        result[0x377] = (byte)0x02;   // PF7  -> F7
        result[0x378] = (byte)0x18;   // PF19 -> F19
        result[0x37F] = (byte)0x0A;   // PF8  -> F8
        result[0x380] = (byte)0x20;   // PF20 -> F20
        result[0x387] = (byte)0x01;   // PF9  -> F9
        result[0x388] = (byte)0x28;   // PF21 -> F21
        result[0x38F] = (byte)0x09;   // PF10 -> F10
        result[0x390] = (byte)0x30;   // PF22 -> F22
        result[0x396] = (byte)0x78;   // PF11 -> F11
        result[0x397] = (byte)0x38;   // PF23 -> F23
        result[0x39E] = (byte)0x07;   // PF12 -> F12
        result[0x39F] = (byte)0x40;   // PF24 -> F24
//
// Now we have scancode translation, we need to be able to select the different
// scancode sets. This means supporting keyboard command 0xF0, which takes one
// parameter byte.
//
// Since the 0xED command ('set LEDS') also takes one parameter byte,
// piggyback on it.
//
// This code is run after receiving and acknowledging the first byte of
// either 0xED or 0xF0. It remembers what the command byte was. Possibly
// it would be safe to store it in R6, but to be on the safe side I'll
// use a memory location.
//
// INIT_MULTI:
        result[0x674] = MOV_R0;     // MOV R0, MULTIBYTE
        result[0x675] = MULTIBYTE;
        result[0x676] = MOV_atR0_A; // MOV @R0, A      ;Remember command
        result[0x677] = JUMP2;      // JMP INIT_MULTI2 ;And continue as for
        result[0x678] = INIT_MULTI2; //                ;setting the LEDs.

// Now divert the inter-page jump to INIT_MULTI2 to the new code.
        result[0xDD] = JUMP6;
        result[0xDE] = INIT_MULTI;
        result[0xF0] = J_INIT_MULTI;    // And execute it when receiving
                                        // scancode 0xF0.

//
// This code is run when the second byte of either a 0xED or 0xF0 command is
// received.
//
// ON_MULTI:
        result[0x679] = MOV_R0;     // MOV R0, MULTIBYTE    ;What command
        result[0x67A] = MULTIBYTE;  //                      ;was it?
        result[0x67B] = XCH_A_atR0; // XCH A, @R0
        result[0x67C] = XRL_A;      // XRL A, 0EDh          ;If 0xED
        result[0x67D] = (byte)0xED; //                      ;then setting LEDs
        result[0x67E] = JZ;         // JZ IS_LEDS
        result[0x67F] = IS_LEDS;
//
// Not setting LEDs, so must be getting or setting the scancode set. Retrieve
// the parameter byte (and, for tidiness, reset MULTIBYTE to 0.
//
        result[0x680] = CLR_A;      // CLR A        ; Set MULTIBYTE to 0
        result[0x681] = XCH_A_atR0; // XCH A, @R0   ; A = parameter byte
        result[0x682] = INC_R0;     // INC R0       ; R0 -> SCANSET
        result[0x683] = JNZ;        // JNZ setting  ; If A is nonzero, select
        result[0x684] = SSC_SETTING;//              ; set in A.
//
// A = 0, so return the current scancode set. When transferring similar data
//       (eg, the keyboard ID, or this function's own parameter byte) IBM's
//       code sets bits in R2. Fortunately bit 0 does not appear to be used,
//       so borrow it to indicate 'returning scancode set'. The main loop
//       will be patched to jump to SSC_RESUME when it sees bit 0 set.
//
//       Also set bits 7 and 2 to disable keyboard scanning while we return
//       the scancode set.
//
        result[0x685] = MOV_R2;     // MOV R2, 85h  ;Set bits 7, 2 and 0
        result[0x686] = (byte)0x85; //
        result[0x687] = JUMP2;      // JMP END_MULTI ;Sends ACK
        result[0x688] = END_MULTI;
//
// Ready to transmit the scancode set. R0 will point at SCANSET.
//
// SSC_RESUME:
        result[0x689] = MOV_A_atR0; // MOV A, @R0       ;Get scancode set.
        result[0x68A] = CALL1;      // CALL TRANSMIT    ;Transmit to host.
        result[0x68B] = TRANSMIT;
        result[0x68C] = JF1;        // JF1 SSC_RETRY    ;If host not listening
        result[0x68D] = SSC_RETRY;  //                  ;then try again later.
// SSC_FINISHED:
        result[0x68E] = MOV_R2;     // MOV R2, 0        ;Success: Reset R2 to 0
        result[0x68F] = (byte)0x00; //                  ;to re-enable keyboard.
// SSC_MAIN:
        result[0x690] = JUMP0;      // JMP KB_MAIN
        result[0x691] = KB_MAIN;
//
// Note: SSC_RETRY really belongs here, but because of the whole hand-assembly
//      thing it's somewhere further down. Sorry. It's very simple:
//
// We are setting the scancode set. A = required set, 1-255.
// Note: We do not bother to check that A is in range; just assume that
// sets 4-255 are all the same as set 2.
//
// SSC_SETTING:
        result[0x692] = MOV_atR0_A; // Store the new scancode set
        result[0x693] = JUMP2;      // JMP END_MULTI
        result[0x694] = END_MULTI;
// SSC_RETRY:
        result[0x695] = MOV_R2;     // MOV R2, 85h  ;Set bits 7, 2 and 0
        result[0x696] = (byte)0x85; //
        result[0x697] = JUMP0;      // JMP KB_MAIN
        result[0x698] = KB_MAIN;

// We are not setting the scancode set: we are setting LEDs. Retrieve the
// parameter and jump to the original code.
//
// IS_LEDS:
        result[0x699] = XCH_A_atR0; // XCH A, @R0    ; Set MULTIBYTE to 0
        result[0x69A] = JUMP2;      // JMP SETLEDS2  ; and A = parameter
        result[0x69B] = SETLEDS2;   //

// And hook ON_MULTI into the 'second byte of set LEDs' code.
        result[0x8D] = JUMP6;       // Second byte of LED -> ON_MULTI
        result[0x8E] = ON_MULTI;


//
// The multi-byte sequences sent by Pause and Break. These have been allocated
// growing down from the top of Page 6.
// BREAK_SC1:
        result[0x6E2] = (byte)0xE0;     // Set 1 sequence for Break:
        result[0x6E3] = (byte)0x46;	// E0 46 E0 C6
        result[0x6E4] = (byte)0xE0;
        result[0x6E5] = (byte)0xC6;
        result[0x6E6] = (byte)0x00;	// and terminating NUL

// PAUSE_SC1:
        result[0x6E7] = (byte)0xE1;     // Set 1 sequence for Pause:
        result[0x6E8] = (byte)0x1D;	// E1 14 77 E1 F0 14 F0 77
        result[0x6E9] = (byte)0x45;
        result[0x6EA] = (byte)0xE1;
        result[0x6EB] = (byte)0x9D;
        result[0x6EC] = (byte)0xC5;
        result[0x6ED] = (byte)0x00;	// and terminating NUL

// BREAK_SC2:
        result[0x6EE] = (byte)0xE0;     // Set 2 sequence for Ctrl+Pause:
        result[0x6EF] = (byte)0x7E; 	// E0 7E E0 F0 7E
        result[0x6F0] = (byte)0xE0;
        result[0x6F1] = (byte)0xF0;
        result[0x6F2] = (byte)0x7E;
        result[0x6F3] = (byte)0x00;	// and terminating NUL

// PAUSE_SC2:
        result[0x6F4] = (byte)0xE1;     // Set sequence for Pause:
        result[0x6F5] = (byte)0x14;	// E1 14 77 E1 F0 14 F0 77
        result[0x6F6] = (byte)0x77;
        result[0x6F7] = (byte)0xE1;
        result[0x6F8] = (byte)0xF0;
        result[0x6F9] = (byte)0x14;
        result[0x6FA] = (byte)0xF0;
        result[0x6FB] = (byte)0x77;
        result[0x6FC] = (byte)0x00;	// and terminating NUL
//
// To complete support for three scancode sets, we need to make sure the
// keyboard powers up in set 2. While we're here, initialise the second 64
// bytes of RAM to all-zeroes.
//
// This code will be run at reset. The IBM code I'm tagging onto uses
// register bank 1, so do the same.
//
// EXTINITRAM:
        result[0x7DA] = MOV_R4;     // MOV R4, 2Ch    ; Default typematic rate
        result[0x7DB] = 0x2C;
        result[0x7DC] = SEL_RB1;    // SEL RB1
        result[0x7DD] = MOV_R0;     // MOV R0, SCANSET
        result[0x7DE] = SCANSET;
        result[0x7DF] = MOV_atR0;   // MOV @R0, 2     ;Set scancode set to 2.
        result[0x7E0] = 0x02;
//
// This code is run either at reset, or when command 0xF6 (Set Defaults) is
// received. It initialises the second 64 bytes (apart from the scancode set)
// to all-zeroes.
//
// EXTRESETRAM:
        result[0x7E1] = SEL_RB1;    // SEL RB1
        result[0x7E2] = MOV_R0;     // MOV R0, 7Eh   ;Last byte to clear
        result[0x7E3] = (byte)0x7E;
        result[0x7E4] = MOV_R3;     // MOV R3, 3Eh   ;Number of bytes to clear
        result[0x7E5] = (byte)0x3E;
        result[0x7E6] = CLR_A;      // CLR A
// EXTRESETLOOP:
        result[0x7E7] = MOV_atR0_A; // MOV @R0, A    ;Write the zeroes
        result[0x7E8] = DEC_R0;     // DEC R0
        result[0x7E9] = DJNZ_R3;    // DJNZ R3, EXTRESETLOOP
        result[0x7EA] = EXTRESETLOOP;
//
// This is for something we'll come to later: Key characteristics in scancode
// set 3. In the 5271 controller these were stored from 0x0A to 0x29. In this
// hybrid firmware we store them from 0x40 to 0x5F.
//
        result[0x7EB] = MOV_R3;     // MOV R3, 20h ; Init 32 bytes of set 3
        result[0x7EC] = (byte)0x20; //             ;state
        result[0x7ED] = MOV_R0;
        result[0x7EE] = (byte)0x5F; // Set 3 keyflags from 40 to 5F
        result[0x7EF] = JUMP7;      // JMP RESETRAM
        result[0x7F0] = IB3LOOP;
//
// Since we've now run into the checksum code, the rest of this initialisation
// has to be put in another free area.
//
// The initial table data will be stored in page 3, from 03C0h to 03DFh. Since
// R3 will range from 01h to 20h, add 0BFh to it to get the right address.
//
// IB3LOOP:
        result[0x714] = MOV_A_R3;    // MOV A, R3
        result[0x715] = ADD_A;
        result[0x716] = (byte)(TABLE_SET3 - 1); // ADD A, 0BFh
        result[0x717] = MOVP3_A_atA; // MOVP3 A, @A ; Byte from init table
        result[0x718] = MOV_atR0_A; // MOV @R0, A
        result[0x719] = DEC_R0;     // DEC R0
        result[0x71A] = DJNZ_R3;    // DJNZ RE3, IB3LOOP
        result[0x71B] = IB3LOOP;
        result[0x71C] = JUMP7;      // JMP RESETRAM  ;And finally chain to
        result[0x71D] = RESETRAM;   //               ;IBM's original reset.

        // Copy the initial key repeat flags from the 5271 ROM to 0x3C0.
        System.arraycopy(rom5271, 0x101, result, 0x3C0,  32);
//
// Now we have the new reset code, hook it in.
//
        result[0x102] = JUMP7;      // Reset leaves the scancode set alone
        result[0x103] = EXTRESETRAM;
        result[0x132] = CALL7;      // Init sets scancodes to set 2
        result[0x133] = EXTINITRAM;

///////////////////////////////////////////////////////////////////////////////
//
// Next, we need the keyboard to support command 0xF2 (identify). Claim to
// be a 122-key terminal emulator board. The ID will be held in page 3
// for ease of configuration, though we could still use the DIP switches
// on P1 if we wanted.
//
// The main loop will call this to send the first ID byte...
//
// SENDID1:
        result[0x69C] = MOV_A;      // MOV A, 0E0h  // First ID byte is at 3E0
        result[0x69D] = TABLE_ID;
        result[0x69E] = JUMP6;      // JMP SENDID
        result[0x69F] = SENDID;
//
// ... and this to send the second ID byte.
//
// SENDID2:
        result[0x6A0] = MOV_A;      // MOV A, 0E1h  // Second ID byte is at 3E1
        result[0x6A1] = (byte)(TABLE_ID + 1);
// SENDID:
        result[0x6A2] = MOVP3_A_atA; // MOVP3 A,@A
        result[0x6A3] = CALL1;      // CALL TRANSMIT
        result[0x6A4] = TRANSMIT;
        result[0x6A5] = JF1;        // JF1 SSC_MAIN
        result[0x6A6] = SSC_MAIN;
//
// As with returning the scancode set above, bits in R2 are used to determine
// which byte we're on. If bit 4 is set, the value we've just sent was the
// second byte, so reset R2 to zero. If not, we've just sent the first byte;
// set bit 4 of R2 and keep going.
//
        result[0x6A7] = MOV_A_R2;   // MOV A, R2
        result[0x6A8] = JB4;        // JB4 SSC_FINISHED  ;Second ID byte sent
        result[0x6A9] = SSC_FINISHED;
        result[0x6AA] = MOV_R2;     // MOV R2, 94h      ;Next ID byte sent is 2nd
        result[0x6AB] = (byte)0x94;
        result[0x6AC] = JUMP0;      // JMP KB_MAIN
        result[0x6AD] = KB_MAIN;
//
// And store the keyboard ID at 0x3E0.
//
        result[0x3E0] = (byte)0xAB;   // ID byte 1
        result[0x3E1] = (byte)0x86;   // ID byte 2
//
// Add code to the keyboard main loop to check bits 0, 4 and 5 of R2, and
// deal with them accordingly. Use some of the space that was originally
// occupied by the scancode translator.
//
        result[0x51] = (byte)0x5D;  // Bit 7 of R7 set: Jump to 0x5D
        result[0x5D] = JB0;         // JB0 J_SSC_RESUME
        result[0x5E] = (byte)0x65;
        result[0x5F] = JB5;         // JB5 J_SENDID1
        result[0x60] = (byte)0x69;
        result[0x61] = JB4;         // JB4 J_SENDID2
        result[0x62] = (byte)0x6B;
        result[0x63] = JUMP0;       // JMP KB_MAIN
        result[0x64] = KB_MAIN;
// J_SSC_RESUME:
        result[0x65] = MOV_R0;      // MOV R0, SCANSET
        result[0x66] = SCANSET;     //
        result[0x67] = JUMP6;       // JMP SSC_RESUME
        result[0x68] = SSC_RESUME;
// J_SENDID_1:
        result[0x69] = JUMP6;
        result[0x6A] = SENDID1;
// J_SENDID_2:
        result[0x6B] = JUMP6;
        result[0x6C] = SENDID2;

// And make command 0xF2 kick off a Send ID sequence.
//
        result[0xF2] = (byte)0x69;  // Command F2 triggers a Send ID

///////////////////////////////////////////////////////////////////////////////
//
// Now for the next biggie: Supporting commands F7-FD, governing how keys
// repeat or not in Set 3. Fortunately, we have the code to do this in the
// 5271 ROM, and we haven't touched Page 4 yet.

// Relocate the key-down code from the 5271 ROM: from 0340-03CF, to 0440-04CF.
// There are rather a lot of fixups, since it makes heavy use of CALL.
//
        System.arraycopy(rom5271, 0x340, result, 0x440,  0x90);

        result[0x450] = ROW_HANDLED;
        result[0x459] = CALL4;      // Relocate all CALL3/JUMP3 instructions
                                    // to CALL4/JUMP4. Replace two subroutines
        result[0x45D] = CALL7;      // MATRIX2SC and PUSHBUFFER with their
        result[0x45E] = MATRIX2SC;  // 5170 counterparts.
        result[0x46A] = CALL4;
        result[0x46E] = CALL7;
        result[0x46F] = MATRIX2SC;
        result[0x472] = JUMP4;
        result[0x474] = CALL4;
        result[0x478] = CALL7;
        result[0x479] = MATRIX2SC;
        result[0x47A] = CALL7;
        result[0x47B] = PUSHBUFFER;
        result[0x47C] = JUMP4;
        result[0x47E] = CALL4;
        result[0x482] = JUMP4;
        result[0x484] = CALL7;
        result[0x485] = TM_REPEAT;

        result[0x486] = CALL7;
        result[0x487] = MATRIX2SC;
        result[0x48D] = JUMP4;
        result[0x49F] = CALL7;
        result[0x4A0] = MATRIX2SC;
        result[0x4A8] = JUMP4;
        result[0x4AA] = CALL7;
        result[0x4AB] = TM_REPEAT;
        result[0x4AC] = JUMP4;
        result[0x4CE] = JUMP4;
        result[0x4B0] = JUMP7;
        result[0x4B1] = ENCODERFAIL;
        result[0x4BD] = 0x40;   // Key characteristics are at 40h, not 0Ah
//
// Use the 5170 decoder or the 5271 decoder, depending on active scanset
//
        result[0x437] = MOV_R0;    // MOV R0, SCANSET
        result[0x438] = SCANSET;
        result[0x439] = MOV_A_atR0;
        result[0x43A] = XRL_A;     // XRL A, 3      ; In set 3?
        result[0x43B] = 0x03;
        result[0x43C] = JZ;        // JZ 0x440      ; Set 3 keydown code
        result[0x43D] = DECODE_3;
        result[0x43E] = JUMP7;     // JMP 0x740     ; Set 1, 2 keydown code
        result[0x43F] = DECODE_2;

        result[0x2A3] = JUMP4;     // JMP 0x437
        result[0x2A4] = DECODE_SW; //


// The 5271 code also has different code in the loop that reads the keyboard
// (among other things, it checks the key flags).
// How necessary this is, I don't know, but copy it anyway. There's not a
// lot of room left in the ROM by now, so the first half goes at the end of
// page 1 and the second half in a gap in page 7.

// J_NO_REPEAT:
        result[0x1CE] = JUMP2;      // For inter-page jumps
        result[0x1CF] = NO_REPEAT;

        result[0x21D] = JUMP1;      // Replace original 5170 repeat code with
        result[0x21E] = (byte)0xD0; // a jump to the following switcher:

        result[0x1D0] = MOV_R0;     // Mov R0, SCANSET
        result[0x1D1] = SCANSET;
        result[0x1D2] = MOV_A_atR0; // MOV A, @R0
        result[0x1D3] = XRL_A;      // XRL A, 3      ;Are we in set 3?
        result[0x1D4] = 0x03;       //
        result[0x1D5] = JZ;         // JZ REPEAT5271 ;Yes: use 5271 repeat code
        result[0x1D6] = REPEAT5271;
        result[0x1D7] = MOV_A_R5;   // MOV A, R5     ;No: replicate original
        result[0x1D8] = ANL_A;      // ANL A, 7Fh    ;5170 code
        result[0x1D9] = 0x7F;
        result[0x1DA] = JUMP2;      // JMP REPEAT5170 ;And jump back to
        result[0x1DB] = REPEAT5170; //                ;where we came from

// The 5271 code from 0x265-0x27F moves to 1E3.
        System.arraycopy(rom5271, 0x265, result, 0x1E3, 27);

        // Fixups to the code:
        result[0x1E7] = J_NO_REPEAT;
        result[0x1E9] = J_NO_REPEAT;
        result[0x1EB] = CALL4;
        result[0x1EC] = SC2MATRIX;
        result[0x1FE] = JUMP7;
        result[0x1FF] = REP5271CONT;

// The 5271 code from 0x280-0x288 moves to 0x71E.
        System.arraycopy(rom5271, 0x280, result, 0x71E, 8);

        // Fixups to the code:
        result[0x721] = (byte)0x1F;
        result[0x723] = CALL4;
        result[0x724] = GET_K_FLAGS;
        result[0x726] = (byte)0x29;
        result[0x727] = JUMP2;
        result[0x728] = NO_REPEAT;

// The 5271 code at 0x292 moves to 0x729.

        System.arraycopy(rom5271, 0x292, result, 0x729, 12);

        // Fixups to the code:
        result[0x72A] = (byte)0x35;
        result[0x732] = (byte)0x35;

        // And rewrite the bit that checks if the key buffer is empty; this
        // is done differently on the 5170 controller.
        result[0x735] = SEL_RB1;    // SEL RB1      ;key buffer registers
        result[0x736] = MOV_A_R2;   // MOV A, R2    ;Bit is 0 if buffer empty
        result[0x737] = JB7;        // JB7 0x73D    ;If not empty, skip
        result[0x738] = 0x3D;
        result[0x739] = SEL_RB0;    // SEL RB0      ;Back to original registers
        result[0x73A] = MOV_A_R6;   // MOV A, R6    ;A = repeating key
        result[0x73B] = CALL7;      // CALL PUSHBUFFER
        result[0x73C] = PUSHBUFFER; //              ; Put scancode in buffer
        result[0x73D] = SEL_RB0;    // SEL RB0
        result[0x73E] = JUMP2;      // JMP 0222
        result[0x73F] = NO_REPEAT;
//
// With the above code present, keys in mode 3 may or may not repeat or send
// break codes. Now we need to support setting their characteristics.
// Grab some subroutines from the 5271 controller ROM, and use the unused
// bits of Page 4 to hold them. This does mean, again, that the first one
// has to be split in half.
//
        System.arraycopy(rom5271, 0x208, result, 0x400, 0x34);
        System.arraycopy(rom5271, 0x23C, result, 0x4D0, 0x20);
        System.arraycopy(rom5271, 0x1fd, result, 0x4E1, 0x0B);
        System.arraycopy(rom5271, 0x1f8, result, 0x4EC, 0x05);

        // Fixups:
        result[0x404] = SKT_F7;
        result[0x408] = SKT_KP;
        result[0x40C] = SKT_BAD;
        result[0x410] = SKT_BAD;
        result[0x413] = SKT_BAD;
        result[0x415] = SKT_BAD;
        result[0x416] = CALL4;
        result[0x417] = SC2MATRIX;
        result[0x41E] = 0x40;   // Key table is now at 0x40, not 0x0A
        result[0x427] = CALL4;
        result[0x428] = RL_MASK;
        result[0x431] = SKT_CONT3;
        result[0x433] = SKT_CONT2;
        result[0x434] = JUMP4;      // This is where we run out of buffer
        result[0x435] = SKT_CONT1;  // and have to resume at SKT_CONT1.

        result[0x4D3] = CALL4;
        result[0x4D4] = RL_MASK;
        result[0x4D8] = DO_SEND_ACK; // Send ACK
        result[0x4DD] = JUMP4;
        result[0x4DE] = SKT_STD;
        result[0x4E0] = DO_SEND_NAK; // Send error

        result[0x4E7] = SC2MAT1;     // Jumps within subroutines
        result[0x4EF] = RL_MASKLOOP;

        result[0x4F1] = MOV_A_R6;   // MOV A, R6       ;Current KB command
        result[0x4F2] = ADD_A;      // ADD A, 8        ;If A >= 0xF8, we are
        result[0x4F3] = 0x08;       //                 ;setting key
        result[0x4F4] = JC;         // JC SET_KTYPE    ;characteristics.
        result[0x4F5] = SET_KTYPE;
        result[0x4F6] = MOV_A_R0;   // MOV A, R0       ;Else go back to what
        result[0x4F7] = XCH_A_R2;   // XCH A, R2       ;we were doing before.
        result[0x4F8] = JUMP0;      // JMP 0x2F
        result[0x4F9] = CHECKLEDS;

        for (n = 0x48F; n < 0x49D; n++)
        {
            result[n] = 0;  // Free. The subroutine here duplicated one in
                            // page 7.
        }

        result[0x2D]  = JUMP4;      // Hook this code into the keyboard loop.
        result[0x2E]  = (byte)0xF1; //
//
// Commands FB / FC / FD set characteristics for a list of keys. Put the
// code for them in spare space at the end of page 5.
//
        result[0xFA]  = (byte)0x6D; // Handle commands FA / FB / FC / FD
        result[0xFB]  = (byte)0x6D;
        result[0xFC]  = (byte)0x6D;
        result[0xFD]  = (byte)0x6D;
        result[0x6D]  = JUMP5;
        result[0x6E]  = (byte)0xE7;
        result[0x6F]  = 0;          // Spare. I thought this might be needed
        result[0x70]  = 0;          // for an inter-page jump, but it wasn't.

        result[0x5E7] = MOV_R2;     // Set bit 6 of R2: Receiving key
        result[0x5E8] = (byte)0xC4; // characteristics.
        result[0x5E9] = JUMP0;
        result[0x5EA] = CLEAR_BUF;  // Flush keyboard queue.

// Commands F7 F8 F9 FA set characteristics of all keys.
// SETALLKEYS:
        result[0x5F3] = MOV_R7;
        result[0x5F4] = 0x20;       // Set 32 keys
        result[0x5F5] = MOV_R0;
        result[0x5F6] = 0x5F;       // From 40-5F
        result[0x5F7] = MOV_atR0_A; // to the value in A.
        result[0x5F8] = DEC_R0;
        result[0x5F9] = DJNZ_R7;
        result[0x5FA] = (byte)0xF7;
        result[0x5FB] = JUMP0;      // Cancel any pending multi-byte transfers
        result[0x5FC] = (byte)0xA3; // (eg: send ID, send scancode set)

        result[0xF7]  = (byte)0xE0; // F7: All keys repeat
        result[0xF8]  = (byte)0xE4; // F8: All keys make/break
        result[0xF9]  = (byte)0xE8; // F9: All keys make only
        result[0xFA]  = (byte)0xE4; // FA: Make it behave like F8.
        result[0xE0]  = MOV_A;      // Set key flags to FF, repeat
        result[0xE1]  = (byte)0xFF;
        result[0xE2]  = JUMP5;
        result[0xE3]  = SETALLKEYS;
        result[0xE4]  = MOV_A;      // Set key flags to 55, make/break
        result[0xE5]  = (byte)0x55;
        result[0xE6]  = JUMP5;
        result[0xE7]  = SETALLKEYS;
        result[0xE8]  = CLR_A;      // Set key flags to 0, make only
        result[0xE9]  = JUMP5;
        result[0xEA]  = SETALLKEYS;

///////////////////////////////////////////////////////////////////////////////
//
// End of patches. Currently-available free areas are:
//
// 06F-070 currently free
// 0EB-0EC currently free
// 1DC-1E6 currently free
// 2EB-2FF currently free
// 3E2-3FF currently free
// 48F-49C currently free
// 4FA-4FC currently free
// 5EB-5F2 currently free
// 6AE-6E1 currently free
//

        checksum(result, 0xDF);
        return result;
    }

    // Constants used in the 2K hybrid ROM:

    // Addresses in data space
    // ~~~~~~~~~~~~~~~~~~~~~~~

// Current scancode set
    private static final byte SCANSET    = 0x7F;
// Current multibyte command
    private static final byte MULTIBYTE  = 0x7E;
    // A flag to indicate that a scancode is being expanded into a multi-byte
    // sequence.
    // For a normal key:         00h, always.
    // For an E0 nn combination: 00h until the E0 is sent, then 80h
    // For a longer byte string: Offset of next byte to send in ROM page 6.
    private static final byte FLAG_E0 = 0x7D;

    // The current scancode being translated (when converting the keyboard's
    // default set 3 codes to set 1 or 2)
    private static final byte CUR_SC  = 0x7C;

// I wanted to keep a bitmap of currently-pressed modifier keys, but so far
// haven't got it to work properly. It would have used these two memory
// locations:
    private static final byte SHIFT_SCRATCH= 0x7B;  // Scratch byte for shift checker
    private static final byte SHIFT_STATE  = 0x7A;  // Bitmap of current shifts
					     // bit 0 => left shift
					     // bit 1 => right shift
					     // bit 2 => left ctrl
					     // bit 3 => right ctrl
					     // bit 4 => left alt
					     // bit 5 => right alt
					     // bit 6 => capslock?



    // Labels in code space
    // ~~~~~~~~~~~~~~~~~~~~
// Page 0
    private static final byte KB_MAIN     = (byte)0x0D;   // Keyboard main loop
    private static final byte DO_SEND_NAK = (byte)0x25;   // Send a NAK.
    private static final byte CHECKLEDS   = (byte)0x2F;   // Check for LED setting
    private static final byte DO_SEND_ACK = (byte)0x41;   // Send an ACK.
    private static final byte DO_SEND_SC  = (byte)0x71;   // Send scancode in A.
    private static final byte SWALLOW_SC  = (byte)0x75;   // Ignore scancode.
    private static final byte DO_SEND_F0  = (byte)0x7E;   // Send scancode 0xF0.
    private static final byte CLEAR_BUF   = (byte)0xA5;   // Clear buffer & continue
    private static final byte CKSUM_P0    = (byte)0xD9;   // Checksum page 0
    private static final byte J_INIT_MULTI= (byte)0xDD;   // Start of multibyte command

// Page 1
    private static final byte ROW_HANDLED = (byte)0x17;   // Keyboard row handled
    private static final byte TRANSMIT    = (byte)0x83;   // Transmit a scancode
    private static final byte J_NO_REPEAT = (byte)0xCE;   // Key is not repeating
    private static final byte REPEAT5271  = (byte)0xE3;   // Key repeat (5271 version)

// Page 2
    private static final byte ADD_CKSUM   = (byte)0x01;   // Add A to ROM checksum
    private static final byte SCAN_KB     = (byte)0x05;   // Scan matrix
    private static final byte REPEAT5170  = (byte)0x20;   // Key repeat (5170 version)
    private static final byte NO_REPEAT   = (byte)0x22;   // No repeat
    private static final byte INIT_MULTI2 = (byte)0xC5;   // Start of multibyte command
    private static final byte SETLEDS2    = (byte)0xCD;   // Start of multibyte command
    private static final byte END_MULTI   = (byte)0xE5;   // End of multi-byte command

// Page 3
    private static final byte TABLE_SET2  = (byte)0x40;   // Set 3 -> 2 key table
    private static final byte TABLE_SET3  = (byte)0xC0;   // Set 3 key characteristics
    private static final byte TABLE_ID    = (byte)0xE0;   // Keyboard ID

// Page 4
    private static final byte SET_KTYPE   = (byte)0x00;   // Set key type
    private static final byte SKT_STD     = (byte)0x16;
    private static final byte DECODE_SW   = (byte)0x37;   // Decoder switch
    private static final byte DECODE_3    = (byte)0x40;   // Decode by set 3 rules
    private static final byte GET_K_FLAGS = (byte)0xB8;   // Get flags for key
    private static final byte SKT_F7      = (byte)0xDB;
    private static final byte SKT_KP      = (byte)0xD9;
    private static final byte SKT_BAD     = (byte)0xDF;
    private static final byte SKT_CONT1   = (byte)0xD0;
    private static final byte SKT_CONT2   = (byte)0xD2;
    private static final byte SKT_CONT3   = (byte)0xD3;
    private static final byte SC2MATRIX   = (byte)0xE1;
    private static final byte SC2MAT1     = (byte)0xE9;
    private static final byte RL_MASKLOOP = (byte)0xEC;
    private static final byte RL_MASK     = (byte)0xEE;

// Page 5
    private static final byte XLT_SET1    = (byte)0x80;   // Translate set 1 scancode
    private static final byte SET1_SKIP   = (byte)0x8A;   // Ignore scancode
    private static final byte SET1_NOT0   = (byte)0x8C;   // Set 1, key is not 0x00
    private static final byte SET1_EXT    = (byte)0x96;   // Set 1, key is extended
    private static final byte SET1_SENDE0 = (byte)0xB3;   // Send scancode 0xE0
    private static final byte SET1_PAUSE  = (byte)0xB5;   // Send Pause sequence
    private static final byte SET1_PAUSE1 = (byte)0xBF;   // Send Pause sequence
    private static final byte SET1_BREAK  = (byte)0xC1;   // Send Break sequence
    private static final byte XLT_SWITCH  = (byte)0xCD;   // Switch translation
    private static final byte J_SET2      = (byte)0xDC;   // Jump to XLT_SET2
    private static final byte J_SET3      = (byte)0xDE;   // Jump to XLT_SET3
    private static final byte SET3_SENDF0 = (byte)0xE5;   // Jump to DO_SENDF0
    private static final byte SETALLKEYS  = (byte)0xF3;   // Set characteristics of all keys

// Page 6
    private static final byte XLT_SET2    = (byte)0x00;   // Translate set 2 scancode
    private static final byte SET2_SKIP   = (byte)0x0A;   // Ignore scancode
    private static final byte SET2_NOT0   = (byte)0x0C;   // Set 2, key is not 0x00
    private static final byte SET2_UP     = (byte)0x16;   // Set 2, key-up code
    private static final byte SET2_EXT    = (byte)0x18;   // Set 2, key is not 0x00
    private static final byte DO_SEND_E0  = (byte)0x36;   // Send scancode 0xE0
    private static final byte SET2_SCAN   = (byte)0x40;   // Scan key matrix
    private static final byte SET2_SENDERR= (byte)0x42;   // Send failed
    private static final byte SET2_SEND_F0= (byte)0x44;   // Send scancode 0xF0
    private static final byte SET2_PAUSE  = (byte)0x46;   // Send Pause sequence
    private static final byte SET2_PAUSE1 = (byte)0x57;   // Send Pause sequence
    private static final byte SET2_BREAK  = (byte)0x68;   // Send Break sequence
    private static final byte BREAK_SC1   = (byte)0xE2;   // Break key sequence (set 1)
    private static final byte PAUSE_SC1   = (byte)0xE7;   // Pause key sequence (set 1)
    private static final byte BREAK_SC2   = (byte)0xEE;   // Break key sequence (set 2)
    private static final byte PAUSE_SC2   = (byte)0xF4;   // Pause key sequence (set 2)
    private static final byte INIT_MULTI  = (byte)0x74;   // Start of 2-byte command
    private static final byte ON_MULTI    = (byte)0x79;   // Second byte of 2-byte cmd
    private static final byte SSC_RESUME  = (byte)0x89;   // Resume sending scanset
    private static final byte SSC_FINISHED= (byte)0x8E;   // Completed sending a multibyte code
    private static final byte SSC_MAIN    = (byte)0x90;   // Jump to KB_MAIN
    private static final byte SSC_SETTING = (byte)0x92;   // Set scanset to A
    private static final byte SSC_RETRY   = (byte)0x95;   // Retry sending scanset
    private static final byte IS_LEDS     = (byte)0x99;   // Set LEDs to @R0
    private static final byte SENDID1     = (byte)0x9C;   // Send ID (first byte)
    private static final byte SENDID2     = (byte)0xA0;   // Send ID (second byte)
    private static final byte SENDID      = (byte)0xA2;   // Send ID (byte at 0x300+A)

// Page 7
    private static final byte XLT_SET3    = (byte)0x00;   // Translate set 3 scancode
    private static final byte SET3_NOT02  = (byte)0x08;   // Set 3, key is not 0x02
    private static final byte SET3_NOT7F  = (byte)0x10;   // Set 3, key is not 0x7F
    private static final byte IB3LOOP     = (byte)0x14;   // Initialise set 3 key flags
    private static final byte REP5271CONT = (byte)0x1E;   // More 5271 repeat code
    private static final byte DECODE_2    = (byte)0x40;   // Decode by set 2 rules
    private static final byte TM_REPEAT   = (byte)0x7B;   // Calculate key delay
    private static final byte ENCODERFAIL = (byte)0x89;   // Report encoder failure
    private static final byte MATRIX2SC   = (byte)0x8F;   // Key number -> set 3 code
    private static final byte PUSHBUFFER  = (byte)0x9B;   // Add scancode to buffer
    private static final byte RESETRAM    = (byte)0xB4;   // Original reset routine
    private static final byte EXTINITRAM  = (byte)0xDA;   // Initialise RAM at reset
    private static final byte EXTRESETRAM = (byte)0xE1;   // Reset RAM to defaults
    private static final byte EXTRESETLOOP= (byte)0xE7;   // New reset routine
    private static final byte CKSUM2K     = (byte)0xF1;   // New checksum code @ 7F1.
    private static final byte CKSUM_P7    = (byte)0xFD;   // Checksum pages 4/5/6/7


    // Selected i8048 opcodes
    private static final byte JUMP0 = (byte)0x04;     // Jump to address in page 0.
    private static final byte JUMP1 = (byte)0x24;     // Jump to address in page 1.
    private static final byte JUMP2 = (byte)0x44;     // Jump to address in page 2.
    private static final byte JUMP3 = (byte)0x64;     // Jump to address in page 3.
    private static final byte JUMP4 = (byte)0x84;     // Jump to address in page 4.
    private static final byte JUMP5 = (byte)0xA4;     // Jump to address in page 5.
    private static final byte JUMP6 = (byte)0xC4;     // Jump to address in page 6.
    private static final byte JUMP7 = (byte)0xE4;     // Jump to address in page 7.

    private static final byte CALL0 = (byte)0x14;     // Call address in page 0.
    private static final byte CALL1 = (byte)0x34;     // Call address in page 1.
    private static final byte CALL2 = (byte)0x54;     // Call address in page 2.
    private static final byte CALL3 = (byte)0x74;     // Call address in page 3.
    private static final byte CALL4 = (byte)0x94;     // Call address in page 4.
    private static final byte CALL5 = (byte)0xB4;     // Call address in page 5.
    private static final byte CALL6 = (byte)0xD4;     // Call address in page 6.
    private static final byte CALL7 = (byte)0xF4;     // Call address in page 7.

    private static final byte JF1   = (byte)0x76;     // Jump if flag 1 set
    private static final byte JNZ   = (byte)0x96;     // Jump if A nonzero
    private static final byte JZ    = (byte)0xC6;     // Jump if A zero
    private static final byte JC    = (byte)0xF6;     // Jump if carry set
    private static final byte JB0   = (byte)0x12;     // Jump if bit 0 of A set
    private static final byte JB1   = (byte)0x32;     // Jump if bit 1 of A set
    private static final byte JB2   = (byte)0x52;     // Jump if bit 2 of A set
    private static final byte JB3   = (byte)0x72;     // Jump if bit 3 of A set
    private static final byte JB4   = (byte)0x92;     // Jump if bit 4 of A set
    private static final byte JB5   = (byte)0xB2;     // Jump if bit 5 of A set
    private static final byte JB6   = (byte)0xD2;     // Jump if bit 6 of A set
    private static final byte JB7   = (byte)0xF2;     // Jump if bit 7 of A set

    private static final byte DJNZ_R0 = (byte)0xE8;   // Decrement R3, jump if nonzero
    private static final byte DJNZ_R1 = (byte)0xE9;   // Decrement R3, jump if nonzero
    private static final byte DJNZ_R2 = (byte)0xEA;   // Decrement R3, jump if nonzero
    private static final byte DJNZ_R3 = (byte)0xEB;   // Decrement R3, jump if nonzero
    private static final byte DJNZ_R4 = (byte)0xEC;   // Decrement R4, jump if nonzero
    private static final byte DJNZ_R5 = (byte)0xED;   // Decrement R5, jump if nonzero
    private static final byte DJNZ_R6 = (byte)0xEE;   // Decrement R6, jump if nonzero
    private static final byte DJNZ_R7 = (byte)0xEF;   // Decrement R7, jump if nonzero

    private static final byte MOVP_A_atA  = (byte)0xA3; // MOVP A, @A
    private static final byte MOVP3_A_atA = (byte)0xE3; // MOVP3 A, @A

    private static final byte INC_atR0    = (byte)0x10; // INC @R0
    private static final byte XCH_A_atR0  = (byte)0x20; // XCH A, @R0
    private static final byte XCH_A_R2    = (byte)0x2A; // XCH A, R2
    private static final byte ORL_A_atR0  = (byte)0x40; // ORL A, @R0
    private static final byte XRL_A_atR0  = (byte)0xD0; // XRL A, @R0

    private static final byte INC_R0 = (byte)0x18;	 // INC R0
    private static final byte DEC_R0 = (byte)0xC8;	 // DEC R0

    private static final byte ADD_A  = (byte)0x03;	 // ADD A,nn
    private static final byte MOV_A  = (byte)0x23;	 // MOV A,nn
    private static final byte ORL_A  = (byte)0x43;	 // ORL A,nn
    private static final byte ANL_A  = (byte)0x53;	 // ANL A,nn
    private static final byte XRL_A  = (byte)0xD3;	 // XRL A,nn
    private static final byte DEC_A  = (byte)0x07;	 // DEC A
    private static final byte INC_A  = (byte)0x17;	 // INC A
    private static final byte CLR_A  = (byte)0x27;	 // CLR A

    private static final byte MOV_atR0 = (byte)0xB0;      // MOV @R0,nn
    private static final byte MOV_R0   = (byte)0xB8;      // MOV R0,nn
    private static final byte MOV_R1   = (byte)0xB9;      // MOV R1,nn
    private static final byte MOV_R2   = (byte)0xBA;      // MOV R2,nn
    private static final byte MOV_R3   = (byte)0xBB;      // MOV R3,nn
    private static final byte MOV_R4   = (byte)0xBC;      // MOV R4,nn
    private static final byte MOV_R5   = (byte)0xBD;      // MOV R5,nn
    private static final byte MOV_R6   = (byte)0xBE;      // MOV R6,nn
    private static final byte MOV_R7   = (byte)0xBF;      // MOV R7,nn

    private static final byte SEL_RB0  = (byte)0xC5;	   // SEL RB0
    private static final byte SEL_RB1  = (byte)0xD5;	   // SEL RB1

    private static final byte MOV_atR0_A = (byte)0xA0;  // MOV @R0, A
    private static final byte MOV_A_atR0 = (byte)0xF0;  // MOV A, @R0
    private static final byte MOV_A_R0   = (byte)0xF8;  // MOV A, R0
    private static final byte MOV_A_R1   = (byte)0xF9;  // MOV A, R1
    private static final byte MOV_A_R2   = (byte)0xFA;  // MOV A, R2
    private static final byte MOV_A_R3   = (byte)0xFB;  // MOV A, R3
    private static final byte MOV_A_R4   = (byte)0xFC;  // MOV A, R4
    private static final byte MOV_A_R5   = (byte)0xFD;  // MOV A, R5
    private static final byte MOV_A_R6   = (byte)0xFE;  // MOV A, R6
    private static final byte MOV_A_R7   = (byte)0xFF;  // MOV A, R7

}
